# Copyright 2019 Siemens AG
# SPDX-License-Identifier: MIT

import logging
import sys
from datetime import date, timedelta

import requests

from fossology.enums import TokenScope
from fossology.exceptions import AuthenticationError, FossologyApiError
from fossology.folders import Folders
from fossology.groups import Groups
from fossology.items import Items
from fossology.jobs import Jobs
from fossology.license import LicenseEndpoint
from fossology.obj import Agents, ApiInfo, HealthInfo, User
from fossology.report import Report
from fossology.search import Search
from fossology.uploads import Uploads
from fossology.users import Users

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def fossology_token(
    url,
    username,
    password,
    token_name,
    token_scope=TokenScope.READ,
    token_expire=None,
    version="v2",
):
    """Generate an API token using username/password

    API endpoint: POST /tokens

    :Example:

    >>> from fossology import fossology_token # doctest: +SKIP
    >>> from fossology.obj import TokenScope # doctest: +SKIP
    >>> token = fossology_token("https://fossology.example.com/repo", "Me", "MyPassword", "MyToken", version="v2") # doctest: +SKIP


    :param url: the URL of the Fossology server
    :param username: name of the user the token will be generated for
    :param password: the password of the user
    :param name: the name of the token
    :param scope: the scope of the token (default: TokenScope.READ)
    :param expire: the expire date of the token, e.g. 2019-12-25 (default: max. 30 days)
    :param version: the version of the API to use (default: "v2")
    :type url: string
    :type username: string
    :type password: string
    :type name: string
    :type scope: TokenScope
    :type expire: string
    :type version: string
    :return: the new token
    :rtype: string
    :raises AuthenticationError: if the username or password is incorrect
    :raises FossologyApiError: if another error occurs
    """
    if version == "v2":
        data = {
            "username": username,
            "password": password,
            "tokenName": token_name,
            "tokenScope": token_scope.value,
        }
    else:
        data = {
            "username": username,
            "password": password,
            "token_name": token_name,
            "token_scope": token_scope.value,
        }
    if token_expire:
        if version == "v2":
            data["tokenExpire"] = token_expire
        else:
            data["token_expire"] = token_expire
    else:
        now = date.today()
        if version == "v2":
            data["tokenExpire"] = str(now + timedelta(days=30))
        else:
            data["token_expire"] = str(now + timedelta(days=30))
    try:
        response = requests.post(url + "/api/" + version + "/tokens", data=data)
        if response.status_code == 201:
            token = response.json()["Authorization"]
            return token.replace("Bearer ", "")
        elif response.status_code == 404:
            description = "Authentication error"
            raise AuthenticationError(description, response)
        else:
            description = "Error while generating new token"
            raise FossologyApiError(description, response)
    except requests.exceptions.ConnectionError as error:
        sys.exit(f"Server {url} does not seem to be running or is unreachable: {error}")


class Fossology(
    Folders, Groups, Items, LicenseEndpoint, Uploads, Jobs, Report, Users, Search
):

    """Main Fossology API class

    Authentication against a running Fossology instance is performed using an API token.

    :Example:

    >>> from fossology import Fossology
    >>> foss = Fossology(FOSS_URL, FOSS_TOKEN) # doctest: +SKIP

    :param url: URL of the Fossology instance
    :param token: The API token generated using the Fossology UI
    :param version: the version of the API to use (default: "v2")
    :type url: str
    :type token: str
    :type version: str
    :raises FossologyApiError: if a REST call failed
    :raises AuthenticationError: if the user couldn't be authenticated
    """

    def __init__(self, url, token, version="v2"):
        self.host = url
        self.token = token
        self.users = list()
        self.folders = list()
        self.api = f"{self.host}/api/{version}"
        self.session = requests.Session()
        self.session.headers.update({"Authorization": f"Bearer {self.token}"})
        self.info = self.get_info()
        self.health = self.get_health()
        self.user = self.get_self()
        self.name = self.user.name
        self.rootFolder = self.detail_folder(self.user.rootFolderId)
        self.folders = self.list_folders()

        logger.info(
            f"Authenticated as {self.user.name} against {self.host} using API version {self.info.version}"
        )

    def get_self(self):
        """Perform the first API request and populate user variables

        API Endpoint: GET /users/self

        :return: the authenticated user's details
        :rtype: User
        :raises FossologyApiError: if the REST call failed
        :raises AuthenticationError: if the user couldn't be found
        """
        response = self.session.get(f"{self.api}/users/self")
        if response.status_code == 200:
            user_agents = None
            user_details = response.json()
            if user_details.get("agents"):
                user_agents = Agents.from_json(user_details["agents"])
            user = User.from_json(user_details)
            user.agents = user_agents
            return user
        else:
            description = "Error while getting details about authenticated user"
            raise FossologyApiError(description, response)

    def close(self):
        self.session.close()

    def get_info(self) -> ApiInfo:
        """Get info from the server

        API endpoint: GET /info

        :return: the API information
        :rtype: ApiInfo
        :raises FossologyApiError: if the REST call failed
        """
        response = self.session.get(f"{self.api}/info")
        if response.status_code == 200:
            return ApiInfo.from_json(response.json())
        else:
            description = "Error while getting API info"
            raise FossologyApiError(description, response)

    def get_health(self) -> ApiInfo:
        """Get health from the server

        API endpoint: GET /health

        :return: the server health information
        :rtype: HealthInfo
        :raises FossologyApiError: if the REST call failed
        """
        response = self.session.get(f"{self.api}/health")
        if response.status_code == 200:
            return HealthInfo.from_json(response.json())
        else:
            description = "Error while getting health info"
            raise FossologyApiError(description, response)



# Copyright 2019 Siemens AG
# SPDX-License-Identifier: MIT

__doc__ = """
        The foss_cli cmdline interface uses the provided REST-API to communicate
        with the Fossology Server.
        Logging is implemented using the standard python logging framework.
        Log Level can be adapted using the -v/-vv option to foss_cli.
        Logging could be sent to console (option --log_to_file) and/or to a log_file
        (option --log_to_file). The name of the log_file (default is .foss_cli.log)
        could be adapted using the option --log_file_name <filename>.
"""
import configparser
import logging
import os
import pathlib
import pprint
import secrets
import sys
from getpass import getpass
from logging.handlers import RotatingFileHandler

import click

from fossology import Fossology, fossology_token
from fossology.enums import AccessLevel, ReportFormat, TokenScope
from fossology.exceptions import FossologyApiError, FossologyUnsupported
from fossology.obj import Folder, Summary

logger = logging.getLogger(__name__)
formatter = logging.Formatter(
    "%(asctime)s - %(levelname)s - [%(filename)s:%(lineno)d] - %(message)s"
)

FOSS_LOGGING_MAP = {0: logging.WARNING, 1: logging.INFO, 2: logging.DEBUG}
MAX_SIZE_OF_LOGFILE = 200000
MAX_NUMBER_OF_LOGFILES = 5

DEFAULT_LOG_FILE_NAME = ".foss_cli.log"
DEFAULT_RESULT_DIR = ".foss_cli_results"
DEFAULT_CONFIG_FILE_NAME = ".foss_cli.ini"

JOB_SPEC = {
    "analysis": {
        "bucket": True,
        "copyright_email_author": True,
        "ecc": True,
        "keyword": True,
        "mime": True,
        "monk": True,
        "nomos": True,
        "ojo": True,
        "package": True,
        "specific_agent": True,
    },
    "decider": {
        "nomos_monk": True,
        "bulk_reused": True,
        "new_scanner": True,
        "ojo_decider": True,
    },
    "reuse": {
        "reuse_upload": 0,
        "reuse_group": 0,
        "reuse_main": True,
        "reuse_enhanced": True,
        "reuse_report": True,
        "reuse_copyright": True,
    },
}


def check_get_folder(ctx: click.Context, folder_name: str):
    """Get a Folder Instance based on the Folders name.

    :param ctx: click context
    :param folder_name: the name of the folder
    :type ctx: click.core.Context
    :type folder_name: str
    :return: the requested Folder
    :rtype: Folder()
    :raises FossologyUnsupported: if folder_name is (not or multiple times) found.
    """
    folder_to_use = None
    foss = ctx.obj["FOSS"]
    if folder_name == "":
        logger.warning(
            "folder_name not specified - upload will be to the Fossology Root folder."
        )
        folder_to_use = foss.rootFolder
    else:
        for a_folder in foss.folders:
            if a_folder.name == folder_name:
                logger.debug(f"Found upload folder {folder_name} with id {a_folder.id}")
                if folder_to_use is None:
                    folder_to_use = a_folder
                else:
                    description = "Multiple Folders with same name are not supported."
                    raise FossologyUnsupported(description)
        if folder_to_use is None:
            description = f"Requested Upload Folder {folder_name} does not exist."
            raise FossologyUnsupported(description)
    assert isinstance(folder_to_use, Folder)
    return folder_to_use


def check_get_report_format(format: str):
    """Check if the  given format is a valid ReportFormat value.

    :param format: name of a ReportFormat
    :type format: str
    :return: ReportFormat
    :rtype: Enum
    """
    for member in ReportFormat:
        if member.value == format:
            return member
    logger.fatal(f"Impossible report format {format}")
    sys.exit(1)


def check_get_access_level(level: str):
    """Check if the given level is a valid AccessLevel value.

    :param level: name of a Access level
    :type:  str
    :return: AccessLevel
    :rtype: Enum
    """
    for member in AccessLevel:
        if member.value == level:
            return member
    logger.fatal(f"Impossible access level {level}")
    sys.exit(1)


def needs_later_initialization_of_foss_instance(ctx: click.Context):
    """Check if later on a Fossology Instance will be created.

    :param ctx: click context
    :type ctx: click.core.Context
    :return: Indicator
    :rtype: bool

    """
    logger.debug(
        f"Function needs_later_initialization_of_foss_instance called {pprint.pformat(ctx.obj)}"
    )
    if ctx.obj["IS_REQUEST_FOR_HELP"] or ctx.obj["IS_REQUEST_FOR_CONFIG"]:
        return False
    return True


def get_newest_upload_of_file(ctx: click.Context, filename: str, folder_name: str):
    """Given a  filename and folder_name return the newest upload if available.

    :param ctx: click Context
    :param filename: str
    :param folder_name: str
    :type ctx: click.core.Context
    :type filename: str
    :type folder_name: str
    :return: Upload Instance or None
    :rtype: Upload or None
    """

    foss = ctx.obj["FOSS"]
    the_uploads, _ = foss.list_uploads(
        folder=folder_name if folder_name else foss.rootFolder,
    )
    found = None
    newest_date = "0000-09-05 13:25:38.079869+00"
    for a_upload in the_uploads:
        if filename.endswith(a_upload.uploadname):
            if a_upload.uploaddate > newest_date:
                newest_date = a_upload.uploaddate
                found = a_upload
    if found:
        the_upload = foss.detail_upload(a_upload.id)
        logger.info(
            f"Can reuse upload for {a_upload.uploadname}. The uploads id is {a_upload.id}."
        )
        assert a_upload.id == the_upload.id
        return the_upload
    else:
        return None


def init_foss(ctx: click.Context):
    """Initialize a Fossology Instance and store it in the context.

    :param ctx: click Context
    :type ctx: click.core.Context
    :raises e: KeyError Bearer TOKEN not set in environment
    :return: foss_instance
    :rtype: Fossology
    """
    logger.debug("INIT FOSS")
    if os.path.exists(DEFAULT_CONFIG_FILE_NAME):
        config = configparser.ConfigParser()
        ctx.obj["CONFIG"] = config
        config.read(DEFAULT_CONFIG_FILE_NAME)
        assert "FOSSOLOGY" in config.sections()
        ctx.obj["TOKEN"] = config["FOSSOLOGY"]["token"]
        ctx.obj["USERNAME"] = config["FOSSOLOGY"]["username"]
        ctx.obj["SERVER"] = config["FOSSOLOGY"]["server_url"]
        logger.debug(
            f"Set server token from configfile {ctx.obj['SERVER']}:{ctx.obj['TOKEN']}"
        )
    else:
        logger.debug("INIT FOSS: No config file found")

    if not ctx.obj["TOKEN"]:
        try:
            ctx.obj["TOKEN"] = os.environ["FOSS_TOKEN"]
        except KeyError as e:
            logger.fatal(
                "No Token provided. Either provide FOSS_TOKEN in environment or use the -t option."
            )
            raise e
    foss = Fossology(ctx.obj["SERVER"], ctx.obj["TOKEN"])  # using new API
    ctx.obj["FOSS"] = foss
    ctx.obj["USER"] = foss.user.name
    logger.debug(f"Logged in as user {foss.user.name}")

    return ctx.obj["FOSS"]


@click.group()
@click.option("--token", "-t", help="The token to be used.")
@click.option(
    "--verbose",
    "-v",
    count=True,
    help="Increase verbosity level (e.g. -v -vv).",
)
@click.option(
    "--log_to_console/--no_log_to_console",
    is_flag=True,
    default=True,  # print logging events >= WARNING by default on console
    show_default=True,
    help="Send logging output to console",
)
@click.option(
    "--log_to_file/--no_log_to_file",
    is_flag=True,
    default=False,
    show_default=True,
    help="Send logging output to file",
)
@click.option(
    "--log_file_name",
    default=DEFAULT_LOG_FILE_NAME,
    show_default=True,
    help="Specify log filename",
)
@click.option(
    "--debug/--no_debug",
    is_flag=True,
    default=False,
    show_default=True,
    help="Send detailed logging output to console.",
)
@click.option(
    "--result_dir",
    default=DEFAULT_RESULT_DIR,
    show_default=True,
    help="Name of the directory where foss_cli writes results.",
)
@click.pass_context
def cli(
    ctx: click.core.Context,
    token: str,
    verbose: int,
    log_to_console: bool,
    log_to_file: bool,
    log_file_name: str,
    debug: bool,
    result_dir: str,
):
    """The foss_cli cmdline.  Multiple -v increase verbosity-level."""
    if log_to_console:
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)
    if not os.path.isdir(result_dir):
        os.mkdir(result_dir)

    if log_to_file:
        logfile_handler = RotatingFileHandler(
            os.path.join(result_dir, log_file_name),
            maxBytes=MAX_SIZE_OF_LOGFILE,
            backupCount=MAX_NUMBER_OF_LOGFILES,
        )
        logfile_handler.setFormatter(formatter)
        logger.addHandler(logfile_handler)
    logger.setLevel(FOSS_LOGGING_MAP.get(verbose, logging.DEBUG))
    assert os.path.isdir(result_dir)
    ctx.obj["VERBOSE"] = verbose
    ctx.obj["TOKEN"] = token
    ctx.obj["DEBUG"] = debug
    ctx.obj["RESULT_DIR"] = result_dir

    if ctx.obj["VERBOSE"] >= 2:
        logger.debug(f"foss_cli called with: {pprint.pformat(sys.argv)}")

    foss_needs_initialization = needs_later_initialization_of_foss_instance(ctx)
    if foss_needs_initialization:
        logger.debug("Initializing Fossology client according to the CLI context")
        foss = init_foss(ctx)  # leaves the foss instance within the ctx dict
    else:
        logger.debug("Initialization of Fossology client is not needed")

    if debug:
        logger.debug("Started in debug mode")
        if foss_needs_initialization:
            logger.debug(
                f"Using API: {pprint.pformat(foss.api)} version {pprint.pformat(foss.info.version)}"
            )
            logger.debug(
                f"Running as user {pprint.pformat(foss.user.name)} on {pprint.pformat(foss.host)}"
            )
        else:
            logger.debug("Fossology client was not initialized")


@cli.command("config")
@click.option(
    "--server",
    default="http://fossology/repo",
    show_default=True,
    help="URL of Fossology server",
)
@click.option(
    "--username",
    default="fossy",
    show_default=True,
    help="Username on Fossology server",
)
@click.option("--password", default="fossy", show_default=True, help="Password")
@click.option(
    "--token_scope",
    default="read",
    show_default=True,
    help="Access scope on Fossology Server (read/write)",
)
@click.option(
    "--interactive/--nointeractive",
    is_flag=True,
    default=True,
    show_default=True,
    help="Get config values via stdin",
)
@click.pass_context
def config(
    ctx: click.core.Context,
    server: str,
    username: str,
    password: str,
    token_scope: str,
    interactive: bool,
):
    """Create a foss_cli config file."""

    if interactive:
        print("Enter the URL to your Fossology server: e.g. http://fossology/repo")
        server = input("Fossology URL: ")
        print(
            "Enter Username and Password: e.g. fossy/fossy (in the default environment)"
        )
        username = input("Username: ")
        password = getpass()
        token_scope = "read"
        while True:
            try:
                print(
                    "Enter a scope for your Fossology token: either 'read' or 'write'"
                )
                token_scope = input("Token scope: ")
                assert token_scope in ["read", "write"]
                break
            except Exception:
                print("Allowed values are 'read' or 'write' (using 'read' as default)")

    logger.warning(
        f"Create a new config for {username} on {server} with scope {token_scope}"
    )

    if token_scope == "read":
        the_token_scope = TokenScope.READ
    else:
        the_token_scope = TokenScope.WRITE

    logger.debug(f"Create token for {username} on {server}")
    token = fossology_token(
        server,
        username,
        password,
        secrets.token_urlsafe(8),  # TOKEN_NAME
        the_token_scope,
    )
    logger.debug(f"Token {token} has been created")

    path_to_cfg_file = pathlib.Path.cwd() / DEFAULT_CONFIG_FILE_NAME

    if path_to_cfg_file.exists():
        logger.info(
            f"Found existing foss_cli config file {path_to_cfg_file}, updating the values..."
        )
        config = configparser.ConfigParser()
        config.read(path_to_cfg_file)
    else:
        logger.info(
            f"foss_cli config file {path_to_cfg_file} not found, creating a new one..."
        )
        config = configparser.ConfigParser()

    config["FOSSOLOGY"] = {
        "SERVER_URL": server,
        "USERNAME": username,
        "TOKEN": token,
    }
    with open(path_to_cfg_file, "w") as fp:
        config.write(fp)

    logger.warning(f"New config has been generated in {path_to_cfg_file}")


@cli.command("log")
@click.option(
    "--log_level",
    default=0,
    show_default=True,
    help="Set the log_level of the message (0,1,2)",
)
@click.option(
    "--message_text",
    default="log message",
    show_default=True,
    help="Text of the log message",
)
@click.pass_context
def log(ctx: click.core.Context, log_level: int, message_text: str):
    """Add a Log Message to the log. If a log message is printed to the log depends
    on the verbosity defined starting the foss_cli (default level 0 /-v level 1/-vv level 2).
    Beeing on global verbosity level 0 only messages of --log_level 2 will be printed.
    Beeing on global verbosity level 1 messages of --log_level 1 and 2 will be printed.
    Beeing on global verbosity level 2 messages of --log_level 0,1,2 will be printed.
    Where the log messages are printed depends on the global configuration for --log_to_console,
    --log_to_file and --log_file_name.
    """

    if log_level == 0:
        logger.debug(message_text)
    elif log_level == 1:
        logger.info(message_text)
    elif log_level == 2:
        logger.warning(message_text)
    else:
        error_text = "Impossible Log Level in Log command."
        logger.fatal(error_text)
        raise click.UsageError(error_text, ctx=ctx)


@cli.command("create_folder")
@click.argument("folder_name")
@click.option("--folder_description", help="Description of the Folder.")
@click.option("--folder_group", help="Name of the Group owning the Folder.")
@click.pass_context
def create_folder(
    ctx: click.core.Context,
    folder_name: str,
    folder_description: str,
    folder_group: str,
):
    """The foss_cli create_folder command."""
    foss = ctx.obj["FOSS"]
    logger.debug(
        f" Try to create folder {folder_name} for group {folder_group} desc: {folder_description}"
    )
    try:
        folder = foss.create_folder(
            foss.rootFolder,
            folder_name,
            description=folder_description,
            group=folder_group,
        )
        logger.debug(
            f"Folder {folder.name} with description {folder.description} created"
        )
    except Exception as e:
        logger.fatal(f"Error creating Folder {folder_name} ", exc_info=True)
        raise e


@cli.command("create_group")
@click.argument("group_name")
@click.pass_context
def create_group(ctx: click.core.Context, group_name: str):
    """The foss_cli create_group command."""
    logger.debug(f"Create group {group_name}")
    foss = ctx.obj["FOSS"]
    try:
        foss.create_group(group_name)
        logger.debug(f" group {group_name} created")
    except FossologyApiError as e:
        if f"Group {group_name} already exists" in e.message:
            logger.debug(f"Group {group_name} already exists.")
        else:
            logger.fatal(f"Error adding group {group_name} ", exc_info=True)
            raise e


@cli.command("upload_file")
@click.argument(
    "upload_file",
    type=click.Path(exists=True),
)
@click.option(
    "--folder_name", default="", show_default=True, help="The name of the upload folder"
)
@click.option(
    "--description", default="", show_default=True, help="The description of the upload"
)
@click.option(
    "--access_level",
    default="public",
    show_default=True,
    help="The access level the upload",
)
@click.option(
    "--reuse_newest_upload/--no_reuse_newest_upload",
    is_flag=True,
    default=False,
    help="Reuse last upload if available.",
)
@click.option(
    "--summary/--no_summary",
    is_flag=True,
    default=False,
    help="Get summary of upload.",
)
@click.pass_context
def upload_file(
    ctx: click.Context,
    upload_file: str,
    folder_name: str,
    description: str,
    access_level: str,
    reuse_newest_upload: bool,
    summary: Summary,
):
    """The foss_cli upload_file command."""

    logger.debug(f"Try to upload file {upload_file}")
    foss = ctx.obj["FOSS"]

    # check/set the requested access level
    the_access_level = check_get_access_level(access_level)

    # check/set the requested folder
    folder_to_use = check_get_folder(ctx, folder_name)

    if reuse_newest_upload:
        the_upload = get_newest_upload_of_file(ctx, upload_file, folder_name)
    else:
        the_upload = None

    if the_upload is None:
        the_upload = foss.upload_file(
            folder_to_use,
            file=upload_file,
            description=description if description else "upload via foss-cli",
            access_level=the_access_level,
            wait_time=10,
        )

    ctx.obj["UPLOAD"] = the_upload

    if summary:
        summary = foss.upload_summary(the_upload)
        if ctx.obj["DEBUG"]:
            logger.debug(
                f"Summary of upload {summary.uploadName} ({summary.id})"
                f"Main license: {summary.mainLicense}"
                f"Unique licenses: {summary.uniqueLicenses}"
                f"Total licenses: {summary.totalLicenses}"
                f"Unique concluded licenses: {summary.uniqueConcludedLicenses}"
                f"Total concluded licenses: {summary.totalConcludedLicenses}"
                f"Files to be cleared: {summary.filesToBeCleared}"
                f"Files cleared: {summary.filesCleared}"
                f"Clearing status: {summary.clearingStatus}"
                f"Copyright count: {summary.copyrightCount}"
                f"Additional info: {summary.additional_info}"
            )


@cli.command("delete_folder")
@click.argument("folder_name")
@click.pass_context
def delete_folder(
    ctx: click.core.Context,
    folder_name: str,
):
    """The foss_cli delete_folder command."""

    logger.debug(f"Try to delete folder {folder_name}")
    foss = ctx.obj["FOSS"]

    folder = None
    for f in foss.list_folders():
        if f.name == folder_name:
            folder = f
            logger.debug(f"Found folder to delete: {folder}")
            break

    if not folder:
        logger.fatal(f"Unable to find folder {folder_name}.")
        ctx.exit(1)

    foss.delete_folder(folder)
    logger.debug(f"Delete command was send to {foss.host} for folder {folder}")


@cli.command("delete_upload")
@click.argument("upload_name")
@click.pass_context
def delete_upload(
    ctx: click.core.Context,
    upload_name: str,
):
    """The foss_cli folder_id command."""

    logger.debug(f"Try to delete upload {upload_name}")
    foss = ctx.obj["FOSS"]

    upload = None
    for u in foss.list_uploads(all_pages=True)[0]:
        if u.uploadname == upload_name:
            upload = u
            logger.debug(f"Found upload to delete: {upload}")
            break

    foss.delete_upload(upload)
    logger.debug(f"Delete command was send to {foss.host} for upload {upload}")


@cli.command("start_workflow")
@click.argument(
    "file_name",
    type=click.Path(exists=True),
)
@click.option(
    "--folder_name",
    default="",
    show_default=True,
    help="The name of the folder to upload to",
)
@click.option(
    "--file_description",
    default="Upload via foss-cli",
    show_default=True,
    help="The description of the upload",
)
@click.option(
    "--dry_run/--no_dry_run",
    is_flag=True,
    default=False,
    show_default=True,
    help="Do not upload but show what would be done. Use -vv to see output.",
)
@click.option(
    "--reuse_newest_upload/--no_reuse_newest_upload",
    is_flag=True,
    default=False,
    show_default=True,
    help="Reuse newest upload if available",
)
@click.option(
    "--reuse_newest_job/--no_reuse_newest_job",
    is_flag=True,
    default=False,
    show_default=True,
    help="Reuse newest scheduled job for the upload if available",
)
@click.option(
    "--report_format",
    default="unifiedreport",
    show_default=True,
    help="The name of the reportformat (dep5, spdx2, spdxtv, readmeoss, unifiedreport)",
)
@click.option(
    "--access_level",
    default="protected",
    show_default=True,
    help="The access level of the upload (private, protected, public)",
)
@click.pass_context
def start_workflow(  # noqa: C901
    ctx: click.core.Context,
    file_name: str,
    file_description: str,
    folder_name: str,
    report_format: str,
    access_level: str,
    reuse_newest_upload: bool,
    reuse_newest_job: bool,
    dry_run: bool,
):
    """The foss_cli start_workflow command."""
    global JOB_SPEC
    logger.debug(f"Try to schedule job for {file_name}")
    foss = ctx.obj["FOSS"]

    # check/set the requested report format
    the_report_format = check_get_report_format(report_format)

    # check/set the requested access level
    the_access_level = check_get_access_level(access_level)

    # check/get the folder to use identified by the provided  folder_name
    folder_to_use = check_get_folder(ctx, folder_name)

    # check/get the foss.upload to use
    if reuse_newest_upload:
        the_upload = get_newest_upload_of_file(ctx, file_name, folder_name)
    else:
        if dry_run:
            logger.warning(
                "Skip upload as dry_run is requested without --reuse_newest_upload"
            )
            the_upload = None
        else:
            logger.debug(f"Initiate new upload for {file_name}")

            the_upload = foss.upload_file(
                folder_to_use,
                file=file_name,
                description=file_description,
                access_level=the_access_level,
                wait_time=10,
            )
            logger.debug(f"Finished upload for {file_name}")

    if the_upload is None:
        logger.fatal(f"Unable to find upload for {file_name}.")
        ctx.exit(1)

    # check/get job correlated with the upload
    job = None
    if reuse_newest_job:
        logger.debug(f"Try to find a scheduled job on {the_upload.uploadname}")
        the_jobs, pages = foss.list_jobs(the_upload)
        newest_date = "0000-09-05 13:25:38.079869+00"
        for the_job in the_jobs:
            if the_job.queueDate > newest_date:
                newest_date = the_job.queueDate
                job = the_job
        if job is None:
            logger.info(f"Upload {the_upload.uploadname} never started a job ")
        else:
            logger.debug(
                f"Can reuse old job on Upload {the_upload.uploadname}: Newest Job id {job.id} is from {job.queueDate} "
            )

    if job is None:  # always true if --no_reuse_newest_job
        job = foss.schedule_jobs(
            folder_to_use if folder_to_use else foss.rootFolder,
            the_upload,
            JOB_SPEC,
            wait=True,  # we wait (default 30 sec) for the job to complete
        )
        logger.debug(f"Scheduled new job {job}")

    # check/get state of job correlated with the upload
    logger.debug(f"job  {job.id}  is in state {job.status} ")
    if job.status == "Processing":
        logger.fatal(
            f"job  {job.id}  is still in state {job.status}: Please try again later with --reuse_newest_upload --reuse_newest_job "
        )
        ctx.exit(1)

    assert job.status == "Completed"

    # trigger generation of report
    report_id = foss.generate_report(the_upload, report_format=the_report_format)
    logger.debug(f"Generated report {report_id}")

    # download report
    content, name = foss.download_report(report_id)
    logger.debug(
        f"Report downloaded: {name}:  content type: {type(content)} len:  {len(content)}."
    )

    destination_file = os.path.join(ctx.obj["RESULT_DIR"], name)
    with open(destination_file, "wb") as fp:
        written = fp.write(content)
        assert written == len(content)
        logger.info(
            f"Report written to file: report_name {name}  written to {destination_file}"
        )


def main():
    d = dict()
    d["IS_REQUEST_FOR_HELP"] = False
    d["IS_REQUEST_FOR_CONFIG"] = False
    for arg in sys.argv[1:]:
        if arg == "--help":
            d["IS_REQUEST_FOR_HELP"] = True
        if arg == "config":
            d["IS_REQUEST_FOR_CONFIG"] = True
    cli(obj=d)  # pragma: no cover


if __name__ == "__main__":
    main()



# mypy: disable-error-code="attr-defined"
# Copyright 2019 Siemens AG
# SPDX-License-Identifier: MIT

import logging
import re
import time
from typing import Tuple

from tenacity import TryAgain, retry, retry_if_exception_type, stop_after_attempt

from fossology.enums import ReportFormat
from fossology.exceptions import AuthorizationError, FossologyApiError
from fossology.obj import Upload

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class Report:
    """Class dedicated to all "report" related endpoints"""

    @retry(retry=retry_if_exception_type(TryAgain), stop=stop_after_attempt(3))
    def generate_report(
        self,
        upload: Upload,
        report_format: ReportFormat | None = None,
        group: str | None = None,
    ):
        """Generate a report for a given upload

        API Endpoint: GET /report

        :param upload: the upload which report will be generated
        :param format: the report format (default: ReportFormat.READMEOSS)
        :param group: the group name to choose while generating the report (default: None)
        :type upload: Upload
        :type format: ReportFormat
        :type group: string
        :return: the report id
        :rtype: int
        :raises FossologyApiError: if the REST call failed
        :raises AuthorizationError: if the REST call is not authorized
        """
        headers = {"uploadId": str(upload.id)}
        if report_format:
            headers["reportFormat"] = report_format.value
        else:
            headers["reportFormat"] = "readmeoss"
        if group:
            headers["groupName"] = group

        response = self.session.get(f"{self.api}/report", headers=headers)

        if response.status_code == 201:
            report_id = re.search("[0-9]*$", response.json()["message"])
            return report_id[0]  # type: ignore

        elif response.status_code == 403:
            description = f"Report generation for upload {upload.id} not authorized"
            raise AuthorizationError(description, response)

        elif response.status_code == 503:
            wait_time = response.headers["Retry-After"]
            logger.debug(f"Retry generate report after {wait_time} seconds")
            time.sleep(int(wait_time))
            raise TryAgain

        else:
            description = f"Report generation for upload {upload.uploadname} failed"
            raise FossologyApiError(description, response)

    @retry(retry=retry_if_exception_type(TryAgain), stop=stop_after_attempt(10))
    def download_report(
        self, report_id: int, group: str | None = None, wait_time: int = 0
    ) -> Tuple[str, str]:
        """Download a report

        API Endpoint: GET /report/{id}

        Get report for a given upload. If the report is not ready wait another ``wait_time`` seconds or look at
        the ``Retry-After`` to determine how long the wait period shall be.

        If ``wait_time`` is 0, the time interval specified by the ``Retry-After`` header is used.

        The function stops trying after **10 attempts**.

        :Example:

        >>> from fossology import Fossology
        >>>
        >>> foss = Fossology(FOSS_URL, FOSS_TOKEN, username) # doctest: +SKIP
        >>>
        >>> # Generate a report for upload 1
        >>> report_id = foss.generate_report(foss.detail_upload(1)) # doctest: +SKIP
        >>> # Wait up to 20 Minutes until the report is ready
        >>> report_content, report_name = foss.download_report(report_id, wait_time=120) # doctest: +SKIP
        >>> with open(report_name, "wb") as report_file:
        ...     report_file.write(report_content) # doctest: +SKIP

        :param report_id: the id of the generated report
        :param group: the group name to choose while downloading a specific report (default: None)
        :param wait_time: use a customized upload wait time instead of Retry-After (in seconds, default: 0)
        :type report_id: int
        :type group: string
        :type wait_time: int
        :return: the report content and the report name
        :rtype: Tuple[str, str]
        :raises FossologyApiError: if the REST call failed
        :raises AuthorizationError: if the REST call is not authorized
        :raises TryAgain: if the report generation times out after 10 retries
        """
        headers = dict()
        if group:
            headers["groupName"] = group

        response = self.session.get(f"{self.api}/report/{report_id}", headers=headers)

        if response.status_code == 200:
            content = response.headers["Content-Disposition"]
            report_name_pattern = "(^attachment; filename=)(\"|')?([^\"|']*)(\"|'$)?"
            report_name = re.match(report_name_pattern, content).group(3)  # type: ignore
            return response.content, report_name

        elif response.status_code == 403:
            description = f"Download of report {report_id} not authorized"
            raise AuthorizationError(description, response)

        elif response.status_code == 503:
            if not wait_time:
                wait_time = response.headers["Retry-After"]
            logger.debug(
                f"Retry GET report {report_id} after {wait_time} seconds: {response.json()['message']}"
            )
            time.sleep(int(wait_time))
            raise TryAgain

        else:
            description = f"Download of report {report_id} failed"
            raise FossologyApiError(description, response)



# mypy: disable-error-code="attr-defined"
# Copyright 2023 Siemens AG
# SPDX-License-Identifier: MIT
import logging

from fossology.enums import SearchTypes
from fossology.exceptions import AuthorizationError, FossologyApiError
from fossology.obj import File, SearchResult, Upload

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def search_headers(
    searchType: SearchTypes = SearchTypes.ALLFILES,
    upload: Upload | None = None,
    filename: str | None = None,
    tag: str | None = None,
    filesizemin: int | None = None,
    filesizemax: int | None = None,
    license: str | None = None,
    copyright: str | None = None,
    group: str | None = None,
    page_size: int = 100,
) -> dict:
    headers = {"searchType": searchType.value}
    if upload:
        headers["uploadId"] = str(upload.id)
    if filename:
        headers["filename"] = filename
    if tag:
        headers["tag"] = tag
    if filesizemin:
        headers["filesizemin"] = filesizemin  # type: ignore
    if filesizemax:
        headers["filesizemax"] = filesizemax  # type: ignore
    if license:
        headers["license"] = license
    if copyright:
        headers["copyright"] = copyright
    if group:
        headers["groupName"] = group
    headers["limit"] = str(page_size)
    return headers


class Search:
    """Class dedicated to all "search" related endpoints"""

    def search(
        self,
        searchType: SearchTypes = SearchTypes.ALLFILES,
        upload: Upload | None = None,
        filename: str | None = None,
        tag: str | None = None,
        filesizemin: int | None = None,
        filesizemax: int | None = None,
        license: str | None = None,
        copyright: str | None = None,
        group: str | None = None,
        page_size: int = 100,
        page: int = 1,
        all_pages: bool = False,
    ) -> tuple[list[SearchResult], int]:
        """Search for a specific file

        API Endpoint: GET /search

        :param searchType: Limit search to: directory, allfiles (default), containers
        :param upload: Limit search to a specific upload
        :param filename: Filename to find, can contain % as wild-card
        :param tag: tag to find
        :param filesizemin: Min filesize in bytes
        :param filesizemax: Max filesize in bytes
        :param license: License search filter
        :param copyright: Copyright search filter
        :param group: the group name to choose while performing search (default: None)
        :param page_size: limit the number of uploads per page (default: 100)
        :param page: the number of the page to fetch uploads from (default: 1)
        :param all_pages: get all uploads (default: False)
        :type searchType: one of SearchTypes Enum
        :type upload: Upload
        :type filename: string
        :type tag: string
        :type filesizemin: int
        :type filesizemax: int
        :type license: string
        :type copyright: string
        :type group: string
        :type page_size: int
        :type page: int
        :type all_pages: boolean
        :return: a tuple containing the list of search results and the total number of pages
        :rtype: tuple[list[SearchResult], int]
        :raises FossologyApiError: if the REST call failed
        :raises AuthorizationError: if the REST call is not authorized
        """
        headers = search_headers(
            searchType,
            upload,
            filename,
            tag,
            filesizemin,
            filesizemax,
            license,
            copyright,
            group,
            page_size,
        )

        results_list = list()
        if all_pages:
            # will be reset after the total number of pages has been retrieved from the API
            x_total_pages = 2
        else:
            x_total_pages = page
        while page <= x_total_pages:
            headers["page"] = str(page)
            response = self.session.get(f"{self.api}/search", headers=headers)

            if response.status_code == 200:
                for result in response.json():
                    results_list.append(SearchResult.from_json(result))

                x_total_pages = int(response.headers.get("X-TOTAL-PAGES", 0))
                if not all_pages or x_total_pages == 0:
                    logger.info(
                        f"Retrieved page {page} of uploads, {x_total_pages} pages are in total available"
                    )
                    return results_list, x_total_pages
                page += 1

            else:
                description = "Unable to get a result with the given search criteria"
                raise FossologyApiError(description, response)

        logger.info(f"Retrieved all {x_total_pages} of search results")
        return results_list, x_total_pages

    def filesearch(
        self,
        filelist: list | None = None,
        group: str | None = None,
    ):
        """Search for files from hash sum

        API Endpoint: POST /filesearch

        The response does not generate Python objects yet, the plain JSON data is simply returned.

        :param filelist: the list of files (or containers) hashes to search for (default: [])
        :param group: the group name to choose while performing search (default: None)
        :type filelist: list
        :return: list of items corresponding to the search criteria
        :type group: string
        :rtype: JSON
        :raises FossologyApiError: if the REST call failed
        :raises AuthorizationError: if the REST call is not authorized
        """
        headers = {}
        if group:
            headers["groupName"] = group

        response = self.session.post(
            f"{self.api}/filesearch", headers=headers, json=filelist
        )

        if response.status_code == 200:
            all_files = []
            for hash_file in response.json():
                if hash_file.get("findings"):
                    all_files.append(File.from_json(hash_file))
                else:
                    return "Unable to get a result with the given filesearch criteria"
            return all_files

        elif response.status_code == 403:
            description = (
                "Not authorized to get a result with the given filesearch criteria"
            )
            raise AuthorizationError(description, response)

        else:
            description = "Unable to get a result with the given filesearch criteria"
            raise FossologyApiError(description, response)



# mypy: disable-error-code="attr-defined"
# Copyright 2019 Siemens AG
# SPDX-License-Identifier: MIT

import json
import logging
from json.decoder import JSONDecodeError
from typing import Tuple
from urllib.parse import quote

from fossology.enums import LicenseType
from fossology.exceptions import FossologyApiError
from fossology.obj import License, Obligation

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def check_empty_response(response) -> bool:
    try:
        message = response.json().get("message")
        if message and message == "Can not exceed total pages: 0":
            return True
    except JSONDecodeError:
        # No JSON response available
        pass
    return False


class LicenseEndpoint:
    """Class dedicated to all "license" related endpoints"""

    def list_licenses(
        self,
        active: bool = False,
        kind: LicenseType = LicenseType.ALL,
        page_size: int = 100,
        page: int = 1,
        all_pages: bool = False,
    ) -> Tuple[list[License], int]:
        """Get all license from the DB

        API Endpoint: GET /license

        :param active: list only active licenses (default: False)
        :param kind: list only licenses from type LicenseType  (default: LicenseType.ALL)
        :param page_size: the maximum number of results per page (default: 100)
        :param page: the number of pages to be retrieved (default: 1)
        :param all_pages: get all licenses (default: False)
        :type active: bool
        :type kind: LicenseType
        :type page_size: int
        :type page: int
        :type all_pages: boolean
        :return: a list of licenses
        :rtype: List[License]
        :raises FossologyApiError: if the REST call failed
        """
        license_list = list()
        headers = {"limit": str(page_size)}
        if active:
            headers["active"] = json.dumps(True)
        if all_pages:
            # will be reset after the total number of pages has been retrieved from the API
            x_total_pages = 2
        else:
            x_total_pages = page
        while page <= x_total_pages:
            headers["page"] = str(page)
            response = self.session.get(
                f"{self.api}/license?kind={kind.value}", headers=headers
            )
            if response.status_code == 200:
                for license in response.json():
                    license_list.append(License.from_json(license))
                x_total_pages = int(response.headers.get("X-TOTAL-PAGES", 0))
                if not all_pages or x_total_pages == 0:
                    logger.info(
                        f"Retrieved page {page} of license, {x_total_pages} pages are in total available"
                    )
                    return license_list, x_total_pages
                page += 1
            else:
                if check_empty_response(response):
                    return license_list, 0
                description = (
                    f"Unable to retrieve the list of licenses from page {page}"
                )
                raise FossologyApiError(description, response)
        logger.info(f"Retrieved all {x_total_pages} pages of licenses")
        return license_list, x_total_pages

    def detail_license(
        self, shortname: str, group: int | None = None
    ) -> Tuple[int, License, list[Obligation]]:
        """Get a license from the DB

        API Endpoint: GET /license/{shortname}

        :param shortname: Short name of the license
        :param group: the group this license belongs to (default: None)
        :type name: str
        :type group: int
        :return: the license id, the license data and the associated obligations
        :rtype: tuple(int, License, list[Obligation])
        :raises FossologyApiError: if the REST call failed
        """
        headers = dict()
        if group:
            headers["groupName"] = group
        response = self.session.get(
            f"{self.api}/license/{quote(shortname)}", headers=headers
        )
        if response.status_code == 200:
            return License.from_json(response.json())
        elif response.status_code == 404:
            description = f"License {shortname} not found"
            raise FossologyApiError(description, response)
        else:
            description = f"Error while getting license {shortname}"
            raise FossologyApiError(description, response)

    def add_license(self, license: License, merge_request: bool = False):
        """Add a new license to the DB

        API Endpoint: POST /license

        License data are added to the request body, here is an example:

        >>> new_license = License(
        ...     "GPL-1.0",
        ...     "GNU General Public License 1.0",
        ...     "Text of the license...",
        ...     "http://www.gnu.org/licenses/gpl-1.0.txt",
        ...     "red",
        ...     "false"
        ... )
        >>> foss.add_license(new_license, merge_request=True) # doctest: +SKIP

        :param license: the license data
        :param merge_request: open a merge request for the license candidate? (default: False)
        :type license: License
        :type merge_request: bool
        :raises FossologyApiError: if the REST call failed
        """
        headers = {"Content-Type": "application/json"}
        license_data = license.to_dict()
        if merge_request:
            license_data["mergeRequest"] = json.dumps(True)
        response = self.session.post(
            f"{self.api}/license", headers=headers, data=json.dumps(license_data)
        )
        if response.status_code == 201:
            logger.info(f"License {license.shortName} has been added to the DB")
        elif response.status_code == 409:
            logger.info(f"License {license.shortName} already exists")
        else:
            description = f"Error while adding new license {license.shortName}"
            raise FossologyApiError(description, response)

    def update_license(
        self,
        shortname: str,
        fullname: str = "",
        text: str = "",
        url: str = "",
        risk: int = 2,
    ):
        """Update a license

        API Endpoint: PATCH /license/{shortname}

        :param shortName: the short name of the license to be updated
        :param fullName: the new fullName of the license (optional)
        :param text: the new text of the license (optional)
        :param url: the new url of the license (optional)
        :param risk: the new risk of the license (default: 2)
        :type shortName: str
        :type fullName: str
        :type text: str
        :type url: str
        :type risk: int
        :raises FossologyApiError: if the REST call failed
        """
        headers = {"Content-Type": "application/json"}
        license_data = {
            "fullName": fullname,
            "text": text,
            "url": url,
            "risk": str(risk),
        }
        response = self.session.patch(
            f"{self.api}/license/{quote(shortname)}",
            data=json.dumps(license_data),
            headers=headers,
        )
        if response.status_code == 200:
            logger.info(f"License {shortname} has been updated")
        else:
            description = f"Unable to update license {shortname}"
            raise FossologyApiError(description, response)



# Copyright 2023 Siemens AG
# SPDX-License-Identifier: MIT

from enum import Enum


class AccessLevel(Enum):
    """Available access levels for uploads:

    PRIVATE
    PROTECTED
    PUBLIC

    """

    PRIVATE = "private"
    PROTECTED = "protected"
    PUBLIC = "public"


class ReportFormat(Enum):
    """Available report format:

    DEP5
    SPDX2
    SPDX2TV
    READMEOSS
    UNIFIEDREPORT

    """

    DEP5 = "dep5"
    SPDX2 = "spdx2"
    SPDX2TV = "spdx2tv"
    READMEOSS = "readmeoss"
    UNIFIEDREPORT = "unifiedreport"


class SearchTypes(Enum):
    """Type of item that can be searched:

    ALLFILES
    CONTAINERS
    DIRECTORY

    """

    ALLFILES = "allfiles"
    CONTAINERS = "containers"
    DIRECTORY = "directory"


class TokenScope(Enum):
    """Scope for API tokens:

    READ: Read only access, limited only to "GET" calls

    WRITE: Read/Write access, required for calls other than "GET"

    """

    READ = "read"
    WRITE = "write"


class ClearingStatus(Enum):
    """Clearing statuses:

    OPEN
    INPROGRESS
    CLOSED
    REJECTED

    """

    OPEN = "Open"
    INPROGRESS = "InProgress"
    CLOSED = "Closed"
    REJECTED = "Rejected"


class JobStatus(Enum):
    """Job statuses:

    COMPLETED
    FAILED
    QUEUED
    PROCESSING

    """

    COMPLETED = "Completed"
    FAILED = "Failed"
    QUEUED = "Queued"
    PROCESSING = "Processing"


class LicenseType(Enum):
    """License types:

    CANDIDATE
    MAIN
    ALL

    """

    CANDIDATE = "candidate"
    MAIN = "main"
    ALL = "all"


class ObligationClass(Enum):
    """Classification of an obligation:

    GREEN
    WHITE
    YELLOW
    RED

    """

    GREEN = "green"
    WHITE = "white"
    YELLOW = "yellow"
    RED = "red"


class MemberPerm(Enum):
    """Group member permissions:

    USER
    ADMIN
    ADVISOR

    """

    USER = 0
    ADMIN = 1
    ADVISOR = 2


class Permission(Enum):
    """Upload or group permissions:

    NONE
    READ_ONLY
    READ_WRITE
    CLEARING_ADMIN
    ADMIN
    """

    NONE = "0"
    READ_ONLY = "1"
    READ_WRITE = "3"
    CLEARING_ADMIN = "5"
    ADMIN = "10"


class ClearingScope(Enum):
    """Scope of the clearing:

    LOCAL
    PACKAGE
    GLOBAL
    """

    LOCAL = "local"
    PACKAGE = "package"
    GLOBAL = "global"


class ClearingType(Enum):
    """Type of the clearing:

    TO_BE_DISCUSSED
    IRRELEVANT
    IDENTIFIED
    DO_NOT_USE
    NON_FUNCTIONAL
    """

    TO_BE_DISCUSSED = "TO_BE_DISCUSSED"
    IRRELEVANT = "IRRELEVANT"
    IDENTIFIED = "IDENTIFIED"
    DO_NOT_USE = "DO_NOT_USE"
    NON_FUNCTIONAL = "NON_FUNCTIONAL"


class PrevNextSelection(Enum):
    """Type of file to be selected for the prev-next endpoint:

    WITHLICENSES
    NOCLEARING
    """

    WITHLICENSES = "withLicenses"
    NOCLEARING = "noClearing"


class CopyrightStatus(Enum):
    """Status of the copyrights:

    ACTIVE
    INACTIVE
    """

    ACTIVE = "active"
    INACTIVE = "inactive"



# mypy: disable-error-code="attr-defined"
# Copyright 2019 Siemens AG
# SPDX-License-Identifier: MIT
import json
import logging
import re
import time
from typing import Tuple

import requests
from tenacity import TryAgain, retry, retry_if_exception_type, stop_after_attempt

from fossology.enums import AccessLevel, ClearingStatus
from fossology.exceptions import AuthorizationError, FossologyApiError
from fossology.obj import (
    Folder,
    Group,
    Permission,
    Summary,
    Upload,
    UploadCopyrights,
    UploadLicenses,
    UploadPermGroups,
    User,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def list_uploads_parameters(
    folder: Folder | None = None,
    recursive: bool = True,
    name: str | None = None,
    status: ClearingStatus | None = None,
    assignee: str | None = None,
    since: str | None = None,
    group: str | None = None,
    limit: str | None = None,
) -> dict:
    """Helper function to list of query parameters for GET /uploads endpoint"""
    date_pattern = re.compile("^[0-9]{4}-[0-9]{2}-[0-9]{2}")
    params = {}
    if folder:
        params["folderId"] = folder.id
    if not recursive:
        params["recursive"] = "false"
    if name:
        params["name"] = name
    if status:
        params["status"] = status.value
    if assignee:
        params["assignee"] = assignee
    if since:
        if not date_pattern.match(since):
            logger.error(
                f"Date format for 'since' query parameter {since} does not match expected format YYYY-MM-DD"
            )
        else:
            params["since"] = since
    if group:
        params["groupName"] = group
    if limit:
        params["limit"] = limit
    return params


class Uploads:
    """Class dedicated to all "uploads" related endpoints"""

    # Retry until the unpack agent is finished
    @retry(retry=retry_if_exception_type(TryAgain), stop=stop_after_attempt(10))
    def detail_upload(
        self, upload_id: int, group: str | None = None, wait_time: int = 0
    ) -> Upload:
        """Get detailed information about an upload

        API Endpoint: GET /uploads/{id}

        Get information about a given upload. If the upload is not ready wait another ``wait_time`` seconds or look at
        the ``Retry-After`` to determine how long the wait period shall be.

        If ``wait_time`` is 0, the time interval specified by the ``Retry-After`` header is used.

        The function stops trying after **10 attempts**.

        :Examples:

        >>> # Wait up to 20 minutes until the upload is ready
        >>> long_upload = detail_upload(1, wait_time=120) # doctest: +SKIP

        >>> # Wait up to 5 minutes until the upload is ready
        >>> long_upload = detail_upload(1, wait_time=30) # doctest: +SKIP

        :param upload_id: the id of the upload
        :param group: the group the upload shall belong to
        :param wait_time: use a customized upload wait time instead of Retry-After (in seconds, default: 0)
        :type upload_id: int
        :type group: string
        :type wait_time: int
        :return: the upload data
        :rtype: Upload
        :raises FossologyApiError: if the REST call failed
        :raises AuthorizationError: if the REST call is not authorized
        :raises TryAgain: if the upload times out after 10 retries
        """
        headers = {}
        if group:
            headers["groupName"] = group
        response = self.session.get(f"{self.api}/uploads/{upload_id}", headers=headers)

        if response.status_code == 200:
            logger.debug(f"Got details for upload {upload_id}")
            return Upload.from_json(response.json())

        elif response.status_code == 403:
            description = f"Getting details for upload {upload_id} is not authorized"
            raise AuthorizationError(description, response)

        elif response.status_code == 503:
            if not wait_time:
                wait_time = response.headers["Retry-After"]
            logger.debug(
                f"Retry GET upload {upload_id} after {wait_time} seconds: {response.json()['message']}"
            )
            time.sleep(int(wait_time))
            raise TryAgain

        else:
            description = f"Error while getting details for upload {upload_id}"
            raise FossologyApiError(description, response)

    def upload_file(
        self,
        folder: Folder,
        file: str | None = None,
        vcs: dict | None = None,
        url: dict | None = None,
        server: dict | None = None,
        description: str | None = None,
        access_level: AccessLevel | None = None,
        apply_global: bool = False,
        ignore_scm: bool = False,
        group: str | None = None,
        wait_time: int = 0,
    ):
        """Upload a package to FOSSology

        API Endpoint: POST /uploads

        Perform a file, VCS or URL upload and get information about the upload using :func:`~fossology.uploads.Uploads.detail_upload` and passing the ``wait_time`` argument.

        See description of :func:`~fossology.uploads.Uploads.detail_upload` to configure how long the client shall wait for the upload to be ready.

        :Example for a file upload:

        >>> from fossology import Fossology
        >>> from fossology.enums import AccessLevel
        >>> foss = Fossology(FOSS_URL, FOSS_TOKEN, username) # doctest: +SKIP
        >>> my_upload = foss.upload_file(
        ...        foss.rootFolder,
        ...        file="my-package.zip",
        ...        description="My product package",
        ...        access_level=AccessLevel.PUBLIC,
        ...    )  # doctest: +SKIP

        :Example for a VCS upload:

        >>> vcs = {
        ...        "vcsType": "git",
        ...        "vcsUrl": "https://github.com/fossology/fossology-python",
        ...        "vcsName": "fossology-python-github-master",
        ...        "vcsUsername": "",
        ...        "vcsPassword": "",
        ...    }
        >>> vcs_upload = foss.upload_file(
        ...        foss.rootFolder,
        ...        vcs=vcs,
        ...        description="Upload from VCS",
        ...        access_level=AccessLevel.PUBLIC,
        ...    )  # doctest: +SKIP

        :Example for a URL upload:

        >>> url = {
        ...        "url": "https://github.com/fossology/fossology-python/archive/master.zip",
        ...        "name": "fossology-python-master.zip",
        ...        "accept": "zip",
        ...        "reject": "",
        ...        "maxRecursionDepth": "1",
        ...    }
        >>> url_upload = foss.upload_file(
        ...        foss.rootFolder,
        ...        url=url,
        ...        description="Upload from URL",
        ...        access_level=AccessLevel.PUBLIC,
        ...    )  # doctest: +SKIP

        :Example for a SERVER upload:

        >>> server = {
        ...        "path": "/tmp/fossology-python",
        ...        "name": "fossology-python",
        ...    }
        >>> server_upload = foss.upload_file(
        ...        foss.rootFolder,
        ...        server=server,
        ...        description="Upload from SERVER",
        ...        access_level=AccessLevel.PUBLIC,
        ...    )  # doctest: +SKIP


        :param folder: the upload Fossology folder
        :param file: the local path of the file to be uploaded
        :param vcs: the VCS specification to upload from an online repository
        :param url: the URL specification to upload from a url
        :param server: the SERVER specification to upload from fossology server
        :param description: description of the upload (default: None)
        :param access_level: access permissions of the upload (default: protected)
        :param apply_global: apply global decisions for current upload (default: False)
        :param ignore_scm: ignore SCM files (Git, SVN, TFS) (default: True)
        :param group: the group name to chose while uploading the file (default: None)
        :param wait_time: use a customized upload wait time instead of Retry-After (in seconds, default: 0)
        :type folder: Folder
        :type file: string
        :type vcs: dict()
        :type url: dict()
        :type server: dict()
        :type description: string
        :type access_level: AccessLevel
        :type apply_global: boolean
        :type ignore_scm: boolean
        :type group: string
        :type wait_time: int
        :return: the upload data
        :rtype: Upload
        :raises FossologyApiError: if the REST call failed
        :raises AuthorizationError: if the REST call is not authorized
        """
        headers = {"folderId": str(folder.id)}
        if description:
            headers["uploadDescription"] = description
        if access_level:
            headers["public"] = access_level.value
        if apply_global:
            headers["applyGlobal"] = "true"
        if ignore_scm:
            headers["ignoreScm"] = "true"
        if group:
            headers["groupName"] = group

        if file:
            headers["uploadType"] = "file"
            with open(file, "rb") as fp:
                files = {"fileInput": fp}
                response = self.session.post(
                    f"{self.api}/uploads", files=files, headers=headers
                )
        elif vcs or url or server:
            data = dict
            if vcs:
                headers["uploadType"] = "vcs"
                data = {"location": vcs}  # type: ignore
            elif url:
                headers["uploadType"] = "url"
                data = {"location": url}  # type: ignore
            elif server:
                headers["uploadType"] = "server"
                data = {"location": server}  # type: ignore
            headers["Content-Type"] = "application/json"
            response = self.session.post(
                f"{self.api}/uploads", data=json.dumps(data), headers=headers
            )
        else:
            logger.info(
                "Neither VCS, or Url or filename option given, not uploading anything"
            )
            return

        if file:
            source = f"{file}"
        elif vcs:
            source = vcs.get("vcsName")  # type: ignore
        elif url:
            source = url.get("name")  # type: ignore
        elif server:
            source = server.get("name")  # type: ignore

        if response.status_code == 201:
            try:
                upload = self.detail_upload(
                    response.json()["message"], group, wait_time
                )
                logger.info(
                    f"Upload {upload.uploadname} ({upload.hash.size}) "
                    f"has been uploaded on {upload.uploaddate}"
                )
                return upload
            except TryAgain:
                description = f"Upload of {source} failed"
                raise FossologyApiError(description, response)

        elif response.status_code == 403:
            description = f"Upload {description} is not authorized"
            raise AuthorizationError(description, response)

        else:
            description = f"Upload {description} could not be performed"
            raise FossologyApiError(description, response)

    @retry(retry=retry_if_exception_type(TryAgain), stop=stop_after_attempt(3))
    def upload_summary(self, upload: Upload, group=None):
        """Get clearing information about an upload

        API Endpoint: GET /uploads/{id}/summary

        :param upload: the upload to gather data from
        :param group: the group name to chose while accessing an upload (default: None)
        :type: Upload
        :type group: string
        :return: the upload summary data
        :rtype: Summary
        :raises FossologyApiError: if the REST call failed
        :raises AuthorizationError: if the REST call is not authorized
        """
        headers = {}
        if group:
            headers["groupName"] = group
        response = self.session.get(
            f"{self.api}/uploads/{upload.id}/summary", headers=headers
        )

        if response.status_code == 200:
            return Summary.from_json(response.json())

        elif response.status_code == 403:
            description = f"Getting summary of upload {upload.id} is not authorized"
            raise AuthorizationError(description, response)

        elif response.status_code == 503:
            logger.debug(
                f"Unpack agent for {upload.uploadname} (id={upload.id}) didn't start yet"
            )
            time.sleep(3)
            raise TryAgain
        else:
            description = f"No summary for upload {upload.uploadname} (id={upload.id})"
            raise FossologyApiError(description, response)

    @retry(retry=retry_if_exception_type(TryAgain), stop=stop_after_attempt(3))
    def upload_licenses(
        self,
        upload: Upload,
        group: str | None = None,
        agent: str | None = None,
        containers: bool = False,
        license: bool = True,
        copyright: bool = False,
    ) -> list[UploadLicenses] | None:
        """Get clearing information about an upload

        API Endpoint: GET /uploads/{id}/licenses

        :param upload: the upload to gather data from
        :param group: the group name to chose while accessing the upload (default: None)
        :param agent: the license agents to use (e.g. "nomos,monk,ninka,ojo,reportImport", default: "nomos")
        :param containers: wether to show containers or not (default: False)
        :param license: wether to expose license matches (default: True)
        :param copyright: wether to expose copyright matches (default: False)
        :type upload: Upload
        :type group: string
        :type agent: string
        :type containers: boolean
        :type license: boolean
        :type copyright: boolean
        :return: the list of UploadLicenses for the specified agent
        :rtype: list of UploadLicenses
        :raises FossologyApiError: if the REST call failed
        :raises AuthorizationError: if the REST call is not authorized
        """
        params = {
            "containers": containers,
            "license": license,
            "copyright": copyright,
        }
        if agent:
            params["agent"] = agent  # type: ignore
        else:
            params["agent"] = agent = "nomos"  # type: ignore

        headers = {}
        if group:
            headers["groupName"] = group
            params["groupName"] = group  # type: ignore

        response = self.session.get(
            f"{self.api}/uploads/{upload.id}/licenses", params=params, headers=headers
        )

        if response.status_code == 200:
            all_licenses = []
            scanned_files = response.json()
            for file_with_findings in scanned_files:
                file_licenses = UploadLicenses.from_json(file_with_findings)
                all_licenses.append(file_licenses)
            return all_licenses

        elif response.status_code == 403:
            description = f"Getting licenses for upload {upload.id} is not authorized"
            raise AuthorizationError(description, response)

        elif response.status_code == 412:
            description = f"The agent {agent} has not been scheduled for upload {upload.uploadname} (id={upload.id})"
            raise FossologyApiError(description, response)

        elif response.status_code == 503:
            logger.debug("The ununpack agent or queried agents have not started yet.")
            time.sleep(3)
            raise TryAgain

        else:
            description = f"API error while returning license findings for upload {upload.uploadname} (id={upload.id})"
            raise FossologyApiError(description, response)

    @retry(retry=retry_if_exception_type(TryAgain), stop=stop_after_attempt(3))
    def upload_copyrights(
        self,
        upload: Upload,
    ):
        """Get copyright matches from an upload

        API Endpoint: GET /uploads/{id}/copyrights

        :param upload: the upload to gather data from
        :type upload: Upload
        :return: the list of copyrights findings
        :rtype: list of Licenses
        :raises FossologyApiError: if the REST call failed
        :raises AuthorizationError: if the REST call is not authorized
        """
        response = self.session.get(f"{self.api}/uploads/{upload.id}/copyrights")

        if response.status_code == 200:
            all_copyrights = []
            for copyright in response.json():
                all_copyrights.append(UploadCopyrights.from_json(copyright))
            return all_copyrights

        elif response.status_code == 403:
            description = f"Getting copyrights for upload {upload.id} is not authorized"
            raise AuthorizationError(description, response)

        elif response.status_code == 412:
            description = f"The agent has not been scheduled for upload {upload.uploadname} (id={upload.id})"
            raise FossologyApiError(description, response)

        elif response.status_code == 503:
            logger.debug("The ununpack agent or queried agents have not started yet.")
            time.sleep(3)
            raise TryAgain

        else:
            description = f"API error while returning copyright findings for upload {upload.uploadname} (id={upload.id})"
            raise FossologyApiError(description, response)

    def delete_upload(self, upload, group=None):
        """Delete an upload

        API Endpoint: DELETE /uploads/{id}

        :param upload: the upload to be deleted
        :param group: the group name to chose while deleting the upload (default: None)
        :type upload: Upload
        :type group: string
        :raises FossologyApiError: if the REST call failed
        :raises AuthorizationError: if the REST call is not authorized
        """
        headers = {}
        if group:
            headers["groupName"] = group
        response = self.session.delete(
            f"{self.api}/uploads/{upload.id}", headers=headers, timeout=5
        )

        if response.status_code == 202:
            logger.info(f"Upload {upload.id} has been scheduled for deletion")

        elif response.status_code == 403:
            description = f"Not authorized to delete upload {upload.id}"
            raise AuthorizationError(description, response)

        else:
            description = f"Unable to delete upload {upload.id}"
            raise FossologyApiError(description, response)

    def list_uploads(
        self,
        folder: Folder | None = None,
        group: str | None = None,
        recursive: bool = True,
        name: str | None = None,
        status: ClearingStatus | None = None,
        assignee: str | None = None,
        since: str | None = None,
        page_size=100,
        page=1,
        all_pages=False,
    ):
        """Get uploads according to filtering criteria (or all available)

        API Endpoint: GET /uploads

        :param folder: only list uploads from the given folder
        :param group: list uploads from a specific group (not only your own uploads) (default: None)
        :param recursive: wether to list uploads from children folders or not (default: True)
        :param name: filter pattern for name and description
        :param status: status of uploads
        :param assignee: user name to which uploads are assigned to or "-me-" or "-unassigned-"
        :param since: uploads since given date in YYYY-MM-DD format
        :param page_size: limit the number of uploads per page (default: 100)
        :param page: the number of the page to fetch uploads from (default: 1)
        :param all_pages: get all uploads (default: False)
        :type folder: Folder
        :type group: string
        :type recursive: boolean
        :type name: str
        :type status: ClearingStatus
        :type assignee: str
        :type since: str
        :type page_size: int
        :type page: int
        :type all_pages: boolean
        :return: a tuple containing the list of uploads and the total number of pages
        :rtype: Tuple(list of Upload, int)
        :raises FossologyApiError: if the REST call failed
        :raises AuthorizationError: if the REST call is not authorized
        """
        headers = {"limit": str(page_size)}
        if group:
            headers["groupName"] = group

        params = list_uploads_parameters(
            folder=folder,
            recursive=recursive,
            name=name,
            status=status,
            assignee=assignee,
            since=since,
            group=group,
            limit=page_size,
        )
        uploads_list = list()
        if all_pages:
            # will be reset after the total number of pages has been retrieved from the API
            x_total_pages = 2
        else:
            x_total_pages = page
        while page <= x_total_pages:
            headers["page"] = str(page)
            params["page"] = str(page)
            response = self.session.get(
                f"{self.api}/uploads", headers=headers, params=params
            )
            if response.status_code == 200:
                for upload in response.json():
                    uploads_list.append(Upload.from_json(upload))
                x_total_pages = int(response.headers.get("X-TOTAL-PAGES", 0))
                if not all_pages or x_total_pages == 0:
                    logger.info(
                        f"Retrieved page {page} of uploads, {x_total_pages} pages are in total available"
                    )
                    return uploads_list, x_total_pages
                page += 1

            elif response.status_code == 403:
                description = "Retrieving list of uploads is not authorized"
                raise AuthorizationError(description, response)

            else:
                description = f"Unable to retrieve the list of uploads from page {page}"
                raise FossologyApiError(description, response)
        logger.info(f"Retrieved all {x_total_pages} of uploads")
        return uploads_list, x_total_pages

    def update_upload(
        self,
        upload: Upload,
        status: ClearingStatus | None = None,
        comment: str = "",
        assignee: User | None = None,
        group: str | None = None,
    ):
        """Update an upload information

        API Endpoint: PATCH /uploads/{id}

        :param upload: the Upload to be updated
        :param status: the new status of the upload (Open, InProgress, Closed, Rejected)
        :param comment: the comment on the status, required for Closed and Rejected states. Ignored for others. (default: empty)
        :param assignee: the user assigned to the upload (default: None)
        :param group: the group name to chose while changing the upload (default: None)
        :type upload: Upload
        :type status: ClearingStatus
        :type assignee: User
        :type comment: string
        :type group: string
        :raises FossologyApiError: if the REST call failed
        :raises AuthorizationError: if the REST call is not authorized
        """
        params = dict()
        headers = dict()
        if status:
            params["status"] = status.value
        if assignee:
            params["assignee"] = assignee.id  # type: ignore
        if group:
            headers["groupName"] = group
        response = self.session.patch(
            f"{self.api}/uploads/{upload.id}",
            headers=headers,
            params=params,
            data=comment,
        )

        if response.status_code == 202:
            logger.info(f"Upload {upload.uploadname} has been updated")

        elif response.status_code == 403:
            description = f"Updating upload {upload.id} is not authorized"
            raise AuthorizationError(description, response)

        else:
            description = f"Unable to update upload {upload.uploadname}."
            raise FossologyApiError(description, response)

    def move_upload(self, upload: Upload, folder: Folder, action: str):
        """Copy or move an upload by id

        API Endpoint: PUT /uploads/{id}

        :param upload: the Upload to be copied or moved in another folder
        :param folder: the destination Folder
        :param action: the action to be performed, 'copy' or 'move'
        :type upload: Upload
        :type folder: Folder
        :type action: str
        :raises FossologyApiError: if the REST call failed
        :raises AuthorizationError: if the REST call is not authorized
        """
        params = {"folderId": str(folder.id), "action": action}
        response = self.session.put(
            f"{self.api}/uploads/{upload.id}", headers=params, params=params
        )

        if response.status_code == 202:
            logger.info(
                f"Upload {upload.uploadname} has been {action} to {folder.name}"
            )

        elif response.status_code == 403:
            description = f"{action} upload {upload.id} is not authorized"
            raise AuthorizationError(description, response)

        else:
            description = (
                f"Unable to {action} upload {upload.uploadname} to {folder.name}"
            )
            raise FossologyApiError(description, response)

    def download_upload(self, upload: Upload) -> Tuple[str, str]:
        """Download an upload by its id

        API Endpoint: GET /uploads/{id}/download

        :param upload: the Upload to be downloaded
        :type upload: Upload
        :return: the upload content and the upload name
        :rtype: Tuple[str, str]
        :raises FossologyApiError: if the REST call failed
        :raises AuthorizationError: if the REST call is not authorized
        """
        response = self.session.get(f"{self.api}/uploads/{upload.id}/download")

        if response.status_code == 200:
            content = response.headers["Content-Disposition"]
            upload_filename_pattern: str = (
                "(^attachment; filename=)(\"|')?([^\"|']*)(\"|'$)?"
            )
            upload_filename = re.match(upload_filename_pattern, content).group(3)  # type: ignore
            return response.content, upload_filename

        elif response.status_code == 403:
            description = f"Not authorized to download upload {upload.id}"
            raise AuthorizationError(description, response)

        else:
            description = f"Unable to download upload {upload.id}"
            raise FossologyApiError(description, response)

    def change_upload_permissions(
        self,
        upload: Upload,
        all_uploads: bool = False,
        group: Group | None = None,
        new_permission: Permission | None = None,
        public_permission: Permission | None = None,
    ):
        """Change the permission of an upload

        API Endpoint: PUT /uploads/{id}/permission

        :param upload: the upload to update
        :param group: the group you want to add or edit permission for this upload
        :param new_permission: the permission for the selected group
        :param public_permission: the public permission for this upload
        :type upload: Upload
        :type all_uploads: boolean (default: False)
        :type group: Group (default: None)
        :type new_permission: Permission (default: None)
        :type public_permission: Permission (default: None)
        :raises FossologyApiError: if the REST call failed
        :raises AuthorizationError: if the REST call is not authorized
        """
        data = {
            "folderId": upload.folderid,
            "allUploadsPermission": all_uploads,
            "groupId": group.id if group else 0,
            "newPermission": new_permission.name.lower() if new_permission else "none",
            "publicPermission": public_permission.name.lower()
            if public_permission
            else "none",
        }
        response: requests.Response = self.session.put(
            f"{self.api}/uploads/{upload.id}/permissions", json=data
        )

        if response.status_code == 202:
            logger.info(
                f"Permissions for upload {upload.uploadname} have been updated."
            )

        elif response.status_code == 400:
            description = f"Permissions for upload {upload.uploadname} not updated."
            raise FossologyApiError(description, response)

        elif response.status_code == 403:
            description = (
                f"Updating permissions for upload {upload.id} is not authorized"
            )
            raise AuthorizationError(description, response)

        elif response.status_code == 404:
            description = f"Upload {upload.id} does not exists."
            raise FossologyApiError(description, response)

        elif response.status_code == 503:
            description = "Scheduler is not running."
            raise FossologyApiError(description, response)

        else:
            description = (
                f"API error while updating permissions for upload {upload.uploadname}."
            )
            raise FossologyApiError(description, response)

    def upload_permissions(
        self,
        upload: Upload,
    ):
        """Get all the groups with their respective permissions for an upload

        API Endpoint: GET /uploads/{id}/perm-groups

        :param upload: the upload to get permission from
        :type upload: Upload
        :raises FossologyApiError: if the REST call failed
        :raises AuthorizationError: if the REST call is not authorized
        """
        response = self.session.get(f"{self.api}/uploads/{upload.id}/perm-groups")
        if response.status_code == 200:
            return UploadPermGroups.from_json(response.json())

        elif response.status_code == 403:
            description = (
                f"Getting permissions for upload {upload.id} is not authorized"
            )
            raise AuthorizationError(description, response)

        elif response.status_code == 404:
            description = f"Upload {upload.id} does not exists."
            raise FossologyApiError(description, response)
        else:
            description = (
                f"API error while getting permissions for upload {upload.uploadname}."
            )
            raise FossologyApiError(description, response)



# Copyright 2019 Siemens AG
# SPDX-License-Identifier: MIT

__doc__ = """
  Allows to run the foss_cli with: python -m fossology
"""
from fossology.foss_cli import main

if __name__ == "__main__":
    main()



# mypy: disable-error-code="attr-defined"
# Copyright 2019 Siemens AG
# SPDX-License-Identifier: MIT
import logging

from fossology.exceptions import FossologyApiError
from fossology.obj import Agents, User

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class Users:
    """Class dedicated to all "users" related endpoints"""

    def detail_user(self, user_id: int):
        """Get details of Fossology user.

        API Endpoint: GET /users/{id}

        :param id: the ID of the user
        :type id: int
        :return: the requested user's details
        :rtype: User
        :raises FossologyApiError: if the REST call failed
        """
        response = self.session.get(f"{self.api}/users/{user_id}")
        if response.status_code == 200:
            user_agents = None
            user_details = response.json()
            if user_details.get("agents"):
                user_agents = Agents.from_json(user_details["agents"])
            user = User.from_json(user_details)
            user.agents = user_agents
            return user
        else:
            description = f"Error while getting details for user {user_id}"
            raise FossologyApiError(description, response)

    def list_users(self):
        """List all users from the Fossology instance

        API Endpoint: GET /users

        :return: the list of users
        :rtype: list of User
        :raises FossologyApiError: if the REST call failed
        """
        response = self.session.get(f"{self.api}/users")
        if response.status_code == 200:
            users_list = list()
            for user in response.json():
                if user.get("name") == "Default User":
                    continue
                if user.get("email"):
                    foss_user = User.from_json(user)
                    agents = user.get("agents")
                    if agents:
                        foss_user.agents = Agents.from_json(agents)
                    users_list.append(foss_user)
            return users_list
        else:
            description = f"Unable to get a list of users from {self.host}"
            raise FossologyApiError(description, response)

    def create_user(self, user_spec: dict):
        """Create a new Fossology user

        API Endpoint: POST /users

        :Example:

        >>> user_spec = {
        ...     "id": 0,
        ...     "name": "New User",
        ...     "description": "A brand new user",
        ...     "email": "new.user@fossology.com",
        ...     "accessLevel": "none",
        ...     "rootFolderId": 0,
        ...     "emailNotification": True,
        ...     "defaultGroup": 0,
        ...     "agents": {
        ...         "bucket": True,
        ...         "copyright_email_author": True,
        ...         "ecc": True,
        ...         "keyword": True,
        ...         "mime": True,
        ...         "monk": True,
        ...         "nomos": True,
        ...         "ojo": True,
        ...         "package": True,
        ...         "reso": True,
        ...         "heritage": True
        ...     },
        ...     "defaultBucketpool": 0,
        ...     "user_pass": "$PASSWORD",
        ...     "defaultVisibility": "public"
        ... }
        >>> foss.create_user(user_spec)  # doctest: +SKIP

        :param user_spec: the user creation data
        :type user_spec: dict
        :raises FossologyApiError: if the REST call failed
        """
        response = self.session.post(f"{self.api}/users", json=user_spec)
        if response.status_code == 201:
            logger.info(
                f"User {user_spec['name']} was created, call list_users() to get more information."
            )
        elif response.status_code == 409:
            logger.info(f"User {user_spec['name']} already exists.")
        else:
            description = f"Error while creating user {user_spec['name']}"
            raise FossologyApiError(description, response)

    def delete_user(self, user):
        """Delete a Fossology user.

        API Endpoint: DELETE /users/{id}

        :param user: the user to be deleted
        :type user: User
        :raises FossologyApiError: if the REST call failed
        """
        response = self.session.delete(f"{self.api}/users/{user.id}")

        if response.status_code == 202:
            return
        else:
            description = f"Error while deleting user {user.name} ({user.id})"
            raise FossologyApiError(description, response)



# mypy: disable-error-code="attr-defined"
# Copyright 2019 Siemens AG
# SPDX-License-Identifier: MIT

import json
from typing import Iterable

from fossology.enums import ClearingScope, ClearingType, Permission


class Agents(object):

    """FOSSology agents.

    Represents the agents currently configured for a given user.

    :param bucket: run bucket agent on every upload
    :param copyright_email_author: run copyright_email_author agent on every upload
    :param ecc: run ecc agent on every upload
    :param keyword: run keyword agent on every upload
    :param mimetype: run mimetype agent on every upload
    :param monk: run monk agent on every upload
    :param nomos: run nomos agent on every upload
    :param ojo: run ojo agent on every upload
    :param package: run package agent on every upload
    :param kwargs: handle any other agent provided by the fossology instance
    :type bucket: boolean
    :type copyright_email_author: boolean
    :type ecc: boolean
    :type keyword: boolean
    :type mimetype: boolean
    :type monk: boolean
    :type nomos: boolean
    :type ojo: boolean
    :type package: boolean
    :type kwargs: key word argument
    """

    def __init__(
        self,
        bucket,
        copyright_email_author,
        ecc,
        keyword,
        mimetype,
        monk,
        nomos,
        ojo,
        package,
        **kwargs,
    ):
        self.bucket = bucket
        self.copyright_email_author = copyright_email_author
        self.ecc = ecc
        self.keyword = keyword
        self.mimetype = mimetype
        self.monk = monk
        self.nomos = nomos
        self.ojo = ojo
        self.package = package
        self.additional_agents = kwargs

    def to_dict(self):
        """Get a directory with the agent configuration

        :return: the agents configured for the current user
        :rtype: dict
        """
        generic_agents = {
            "bucket": self.bucket,
            "copyright_email_author": self.copyright_email_author,
            "ecc": self.ecc,
            "keyword": self.keyword,
            "mimetype": self.mimetype,
            "monk": self.monk,
            "nomos": self.nomos,
            "ojo": self.ojo,
            "package": self.package,
        }
        try:
            agents = {**generic_agents, **self.additional_agents}
        except AttributeError:
            agents = generic_agents

        return agents

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)

    def to_json(self):
        """Get a JSON object with the agent configuration

        :return: the agents configured for the current user
        :rtype: JSON
        """
        return json.dumps(self.to_dict())


class User(object):

    """FOSSology user.

    Represents the user currently authenticated against the FOSSology server.

    :param id: the user ID
    :param name: the user name
    :param description: further information about the user
    :param email: the user's email
    :param accessLevel: TBD
    :param rootFolderId: the ID of the user's root folder
    :param emailNotification: are email notifications configured for that user?
    :param agents: the default agent configuration for the user
    :param kwargs: handle any other user information provided by the fossology instance
    :type id: int
    :type name: string
    :type description: string
    :type email: string
    :type accessLevel: string
    :type rootFolderId: int
    :type emailNotification: boolean
    :type agents: Agents
    :type kwargs: key word argument
    """

    def __init__(
        self,
        id: int,
        name: str,
        description: str,
        email: str | None = None,
        accessLevel: int | None = None,
        rootFolderId: int | None = None,
        emailNotification: str | None = None,
        default_group: str | None = None,
        agents: dict | None = None,
        **kwargs: dict,
    ):
        self.id = id
        self.name = name
        self.description = description
        self.email = email
        self.accessLevel = accessLevel
        self.rootFolderId = rootFolderId
        self.emailNotification = emailNotification
        self.default_group = default_group
        self.agents = agents
        self.additional_info = kwargs

    def __str__(self):
        return (
            f"User {self.description} ({self.id}), {self.email}, "
            f"access level {self.accessLevel}, "
            f"root folder {self.rootFolderId}, "
            f"default group {self.default_group}"
        )

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)


class UserGroupMember(object):

    """FOSSology group member.

    Represents a member of a group.

    :param user: the user data structure of the member
    :param group_perm: the permission of the user in the group (0: User, 1: Admin, 2: Advisor)
    :param kwargs: handle any other folder information provided by the fossology instance
    :type user: User
    :type group_perm: int
    :type kwargs: key word argument
    """

    def __init__(
        self,
        user: User,
        group_perm: int,
        **kwargs: dict,
    ):
        self.user = User.from_json(user)
        self.group_perm = group_perm
        self.additional_info = kwargs

    def __str__(self):
        return f"Member {self.user.name} ({self.user.id}) has permission {self.group_perm}."

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)


class Folder(object):

    """FOSSology folder.

    Represents a FOSSology folder.

    :param id: the ID of the folder
    :param name: the name of the folder
    :param description: further information about the folder
    :param parent: the ID of the parent folder
    :param kwargs: handle any other folder information provided by the fossology instance
    :type id: int
    :type name: string
    :type description: string
    :type parent: int
    :type kwargs: key word argument
    """

    def __init__(self, id, name, description, parent, **kwargs):
        self.id = id
        self.name = name
        self.description = description
        self.parent = parent
        self.additional_info = kwargs

    def __str__(self):
        return (
            f"{self.name} ({self.id}), '{self.description}', "
            f"parent folder id = {self.parent}"
        )

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)


class Findings(object):

    """FOSSology license findings.

    Represents FOSSology license findings.

    :param scanner: the list of licenses found by the specified scanners
    :param conclusion: the concluded license by user of for a package
    :param copyright: the copyrights found in the package
    :param kwargs: handle any other finding information provided by the fossology instance
    :type scanner: list
    :type conclusion: list
    :type copyright: list
    :type kwargs: key word argument
    """

    def __init__(
        self,
        scanner: list,
        conclusion: list | None = None,
        copyright: list | None = None,
        **kwargs,
    ):
        self.scanner = scanner
        self.conclusion = conclusion
        self.copyright = copyright
        self.additional_info = kwargs

    def __str__(self):
        return (
            f"Licenses found by scanners: {self.scanner}, concluded licenses: {self.conclusion}, "
            f"{len(self.copyright)} copyrights"
        )

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)


class Group(object):

    """FOSSology group.

    Represents a FOSSology group.

    :param id: the ID of the group
    :param name: the name of the group
    :param kwargs: handle any other folder information provided by the fossology instance
    :type id: int
    :type name: string
    :type kwargs: key word argument
    """

    def __init__(self, id, name, **kwargs):
        self.id = id
        self.name = name
        self.additional_info = kwargs

    def __str__(self):
        return f"Group {self.name} ({self.id})"

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)


class PermGroups(object):
    """GroupIds with their respective permissions for a upload

    Represents the group permissions for a FOSSology upload.

    :param perm: the permission
    :param group_pk: the id of the group
    :param group_name: the name of the group
    :type perm: str
    :type group_pk: str
    :type group_name: str
    """

    def __init__(self, perm: str, group_pk: str, group_name: str):
        self.perm = Permission(perm)
        self.group_pk = group_pk
        self.group_name = group_name

    def __str__(self):
        return f"Group {self.group_name} ({self.group_pk}) with {self.perm.name} permission"

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)


class UploadPermGroups(object):
    """Upload permissions

    Represents the permissions for a FOSSology upload.

    :param publicPerm: the public permission of the group
    :param permGroups: array of permGroup objects for the upload
    :param kwargs: handle any other folder information provided by the fossology instance
    :type publicPerm: str
    :type permGroups: array
    :type kwargs: key word argument
    """

    def __init__(self, publicPerm: str, permGroups: list, **kwargs):
        self.publicPerm = Permission(publicPerm)
        self.permGroups = list()
        for perm in permGroups:
            self.permGroups.append(PermGroups.from_json(perm))
        self.additional_info = kwargs

    def __str__(self):
        return f"Upload with {self.publicPerm.name} permission and permissions for groups {self.permGroups}"

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)


class License(object):

    """FOSSology license.

    Represents a FOSSology license.

    :param shortName: the short name of the license
    :param fullName: the full name of the license
    :param text: the text of the license
    :param url: URL of the license text
    :param risk: the risk level of the license
    :param isCandidate: is the license a candidate?
    :param kwargs: handle any other folder information provided by the fossology instance
    :type shortName: string
    :type fullName: string
    :type text: string
    :type url: string
    :type risk: int
    :type isCandidate: bool
    :type kwargs: key word argument
    """

    def __init__(
        self, shortName, fullName, text, url, risk, isCandidate, id=None, **kwargs
    ):
        self.id = id
        self.shortName = shortName
        self.fullName = fullName
        self.text = text
        self.url = url
        self.risk = risk
        self.isCandidate = isCandidate
        self.additional_info = kwargs

    def __str__(self):
        license_type = "License"
        if self.isCandidate:
            license_type = "Candidate license"
        return f"{license_type} {self.fullName} - {self.shortName} ({self.id}) with risk level {self.risk}"

    def to_dict(self):
        """Get a directory with the license data

        :return: the license data
        :rtype: dict
        """
        return {
            "shortName": self.shortName,
            "fullName": self.fullName,
            "text": self.text,
            "url": self.url,
            "risk": self.risk,
            "isCandidate": self.isCandidate,
        }

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)

    def to_json(self) -> str:
        """Get a JSON object with the license data

        :return: the license data
        :rtype: JSON
        """
        return json.dumps(self.to_dict())


class Obligation(object):

    """FOSSology license obligation.

    Represents a FOSSology license obligation.

    :param id: the ID of the obligation
    :param topic: the topic of the obligation
    :param type: the type of the obligation
    :param text: the text of the obligation
    :param classification: level of attention it should raise in the clearing process
    :param comment: comment for the obligation
    :param kwargs: handle any other folder information provided by the fossology instance
    :type id: int
    :type topic: string
    :type type: string
    :type text: string
    :type classification: string
    :type comment: string
    :type kwargs: key word argument
    """

    def __init__(self, id, topic, type, text, classification, comment, **kwargs):
        self.id = id
        self.topic = topic
        self.type = type
        self.text = text
        self.classification = classification
        self.comment = comment
        self.additional_info = kwargs

    def __str__(self):
        return f"Obligation {self.topic}, {self.type} ({self.id}) is classified {self.classification}"

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)


class Hash(object):

    """FOSSology hash.

    Represents a FOSSology file hash values.

    :param sha1: the SHA1 hash sum of the file
    :param md5: the MDA check sum of the file
    :param sha256: the SHA256 hash sum of the file
    :param size: the size of the file in bytes
    :param kwargs: handle any other hash information provided by the fossology instance
    :type sha1: string
    :type md5: string
    :type sha256: string
    :type size: int
    :type kwargs: key word argument
    """

    def __init__(
        self,
        sha1,
        md5,
        sha256,
        size,
        **kwargs,
    ):
        self.sha1 = sha1
        self.md5 = md5
        self.sha256 = sha256
        self.size = size
        self.additional_info = kwargs

    def __str__(self):
        return f"File SHA1: {self.sha1} MD5 {self.md5} SH256 {self.sha256} Size {self.size}B"

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)


class File(object):

    """FOSSology file response from filesearch.

    Represents a FOSSology filesearch response.

    :param hash: the hash information of the file
    :param findings: the license findings in that file
    :param kwargs: handle any other license information provided by the fossology instance
    :type hash: Hash
    :type findings: Findings
    :type kwargs: key word argument
    """

    def __init__(
        self,
        hash,
        findings,
        **kwargs,
    ):
        self.hash = Hash.from_json(hash)
        self.findings = Findings.from_json(findings)
        self.additional_info = kwargs

    def __str__(self):
        if self.findings.conclusion:
            return f"File with SHA1 {self.hash.sha1} has {len(self.findings.conclusion)} concluded licenses"
        else:
            return f"File with SHA1 {self.hash.sha1} doesn't have any concluded license yet"

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)


class FileInfo(object):

    """FOSSology file info response.

    Represents a FOSSology file info response.

    :param view_info: view info of the file
    :param meta_info: meta info of the file
    :param package_info: package info of the file
    :param tag_info: tag info of the file
    :param reuse_info: reuse info of the file
    :param kwargs: handle any other license information provided by the fossology instance
    :type view_info: Object
    :type meta_info: Object
    :type package_info: Object
    :type tag_info: Object
    :type reuse_info: Object
    :type kwargs: key word argument
    """

    def __init__(
        self,
        view_info,
        meta_info,
        package_info,
        tag_info,
        reuse_info,
        **kwargs,
    ):
        self.view_info = view_info
        self.meta_info = meta_info
        self.package_info = package_info
        self.tag_info = tag_info
        self.reuse_info = reuse_info
        self.additional_info = kwargs

    def __str__(self):
        return f"File view {self.view_info}"

    @classmethod
    def from_json(cls, json_dict):
        for key in ("viewInfo", "metaInfo", "packageInfo", "tagInfo", "reuseInfo"):
            try:
                json_dict[key.replace("I", "_i")] = json_dict[key]
                del json_dict[key]
            except KeyError:
                pass

        return cls(**json_dict)


class Upload(object):

    """FOSSology upload.

    Represents a FOSSology upload.

    :param folderid: the ID of the upload folder
    :param foldername: the name of the upload folder
    :param id: the ID of the upload
    :param description: further information about the upload
    :param uploadname: the name of the upload (default: the name of the upload file)
    :param uploaddate: the date of the upload
    :param assignee: the user who is assigned to the upload
    :param assigneeDate: the date of the assignment
    :param closingDate: the date of the closing
    :param hash: the hash data of the uploaded file
    :param kwargs: handle any other upload information provided by the fossology instance
    :type folderid: int
    :type foldername: string
    :type id: int
    :type description: string
    :type uploadname: string
    :type uploaddate: string
    :type assignee: string
    :type assigneeDate: string
    :type closingDate: string
    :type hash: Hash
    :type kwargs: key word argument
    """

    def __init__(
        self,
        folderid,
        foldername,
        id,
        description,
        uploadname,
        uploaddate,
        assignee=None,
        assigneeDate=None,
        closingDate=None,
        hash=None,
        **kwargs,
    ):
        self.folderid = folderid
        self.foldername = foldername
        self.id = id
        self.description = description
        self.uploadname = uploadname
        self.uploaddate = uploaddate
        self.assignee = (assignee,)
        self.assigneeDate = (assigneeDate,)
        self.closeDate = (closingDate,)
        self.hash = Hash.from_json(hash)
        self.additional_info = kwargs

    def __str__(self):
        return (
            f"Upload '{self.uploadname}' ({self.id}, {self.hash.size}B, {self.hash.sha1}) "
            f"in folder {self.foldername} ({self.folderid})"
        )

    @classmethod
    def from_json(cls, json_dict):
        for key in ("folderId", "folderName", "uploadName", "uploadDate"):
            try:
                json_dict[key.lower()] = json_dict[key]
                del json_dict[key]
            except KeyError:
                pass
        return cls(**json_dict)


class UploadCopyrights(object):

    """Copyright findings in a FOSSology upload

    Represents copyright matches of a FOSSology upload.

    :param copyright: the copyright
    :param filePath: relative file path
    :param kwargs: handle any other information provided by the FOSSology instance
    :type copyright: str
    :type filePath: list
    :type kwargs: key word argument
    """

    def __init__(
        self,
        copyright: str,
        filePath: list[str],
        **kwargs,
    ):
        self.copyright = copyright
        self.filepath = list()
        for path in filePath:
            self.filepath.append(path)
        self.additional_info = kwargs

    def __str__(self):
        return f"Copyright {self.copyright} was found in {len(self.filepath)} files."

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)


class UploadLicenses(object):

    """FOSSology upload licenses.

    Represents licenses and copyright matches of a FOSSology upload.

    :param filePath: relative file path
    :param findings: the licenses and copyrights findings
    :param kwargs: handle any other information provided by the fossology instance
    :type filePath: str
    :type findings: Findings
    :type kwargs: key word argument
    """

    def __init__(
        self,
        filePath: str,
        findings: dict,
        **kwargs,
    ):
        self.filepath = filePath
        self.findings = Findings.from_json(findings)
        self.additional_info = kwargs

    def __str__(self):
        return f"File {self.filepath} has {len(self.findings.conclusion)} license and {len(self.findings.copyright)}matches"

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)


class Summary(object):

    """FOSSology upload summary.

    Represents a FOSSology upload summary.

    :param id: the ID of the upload
    :param uploadName: the name of the upload
    :param mainLicense: the main upload license
    :param uniqueLicenses: the number of unique licenses
    :param totalLicenses: the total number of licenses
    :param uniqueConcludedLicenses: the number of unique concluded licenses
    :param totalConcludedLicenses: the total number of concluded licenses
    :param filesToBeCleared: the number of remaining files to be cleared
    :param filesCleared: the number of files already cleared
    :param clearingStatus: the clearing status
    :param copyrightCount: the number of copyrights found
    :param kwargs: handle any other summary information provided by the fossology instance
    :type id: int
    :type uploadName: string
    :type mainLicense: string
    :type uniqueLicenses: int
    :type totalLicenses: int
    :type uniqueConcludedLicenses: int
    :type totalConcludedLicenses: int
    :type filesToBeCleared: int
    :type filesCleared: int
    :type clearingStatus: string
    :type copyrightCount: int
    :type kwargs: key word argument
    """

    def __init__(
        self,
        id,
        uploadName,
        mainLicense,
        uniqueLicenses,
        totalLicenses,
        uniqueConcludedLicenses,
        totalConcludedLicenses,
        filesToBeCleared,
        filesCleared,
        clearingStatus,
        copyrightCount,
        **kwargs,
    ):
        self.id = id
        self.uploadName = uploadName
        self.mainLicense = mainLicense
        self.uniqueLicenses = uniqueLicenses
        self.totalLicenses = totalLicenses
        self.uniqueConcludedLicenses = uniqueConcludedLicenses
        self.totalConcludedLicenses = totalConcludedLicenses
        self.filesToBeCleared = filesToBeCleared
        self.filesCleared = filesCleared
        self.clearingStatus = clearingStatus
        self.copyrightCount = copyrightCount
        self.additional_info = kwargs

    def __str__(self):
        return (
            f"Clearing status for '{self.uploadName}' is '{self.clearingStatus}',"
            f" main license = {self.mainLicense}"
        )

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)


class Job(object):

    """FOSSology job.

    Represents a FOSSology job.

    :param id: the ID of the job
    :param name: the name of the job
    :param queueDate: the start date of the job
    :param uploadId: the ID of the upload
    :param userId: the ID of the user wo started the job
    :param groupId: the ID of the job's group
    :param eta: the estimated termination time
    :param status: the status of the job
    :param kwargs: handle any other job information provided by the fossology instance
    :type id: int
    :type name: string
    :type queueDate: string
    :type uploadId: int
    :type userId: int
    :type groupId: int
    :type eta: string
    :type status: string
    :type kwargs: key word argument
    """

    def __init__(
        self, id, name, queueDate, uploadId, userId, groupId, eta, status, **kwargs
    ):
        self.id = id
        self.name = name
        self.queueDate = queueDate
        self.uploadId = uploadId
        self.userId = userId
        self.groupId = groupId
        self.eta = eta
        self.status = status
        self.additional_info = kwargs

    def __str__(self):
        return (
            f"Job '{self.name}' ({self.id}) queued on {self.queueDate} "
            f"(Status: {self.status} ETA: {self.eta})"
        )

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)


class ApiLicense(object):

    """FOSSology API License.

    :param name: name of the API license
    :param url: link to the license text
    :type name: string
    :type url: string
    """

    def __init__(self, name, url):
        self.name = name
        self.url = url

    def __str__(self):
        return f"API license '{self.name}' ({self.url})"

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)


class FossologyServer(object):

    """FOSSology server info.

    :param version: version of the FOSSology server (e.g. 4.0.0)
    :param branchName: branch deployed on the FOSSology server
    :param commitHash: hash of commit deployed on the FOSSology server
    :param commitDate: date of commit deployed on the FOSSology server in ISO8601 format
    :param buildDate: date on which packages were built in ISO8601 format
    :type version: string
    :type branchName: string
    :type commitHash: string
    :type commitDate: string
    :type buildDate: string
    """

    def __init__(self, version, branchName, commitHash, commitDate, buildDate):
        self.version = version
        self.branchName = branchName
        self.commitHash = commitHash
        self.commitDate = commitDate
        self.buildDate = buildDate

    def __str__(self):
        return f"Fossology server version {self.version} (branch {self.branchName} - {self.commitHash})"

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)


class ApiInfo(object):

    """FOSSology API info.

    Represents the info endpoint of FOSSology API.

    :param name: the name of the API service
    :param description: additional information
    :param version: current API version
    :param security: security methods allowed
    :param contact: email contact from the API documentation
    :param license: licensing of the API
    :param fossology: information about FOSSology server
    :type name: string
    :type description: string
    :type version: string
    :type security: list
    :type contact: string
    :type license: ApiLicense object
    :type fossology: FossologyServer object
    :type kwargs: key word argument
    """

    def __init__(
        self,
        name,
        description,
        version,
        security,
        contact,
        license,
        fossology,
        **kwargs,
    ):
        self.name = name
        self.description = description
        self.version = version
        self.security = security
        self.contact = contact
        self.license = ApiLicense.from_json(license)
        self.fossology = FossologyServer.from_json(fossology)
        self.additional_info = kwargs

    def __str__(self):
        return f"{self.name} is deployed with version {self.version}"

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)


class Status(object):

    """FOSSology server status

    Represent the status of FOSSology sub-systems

    :param status: the status of the sub-system (OK, ERROR)
    :type status: string
    """

    def __init__(self, status):
        self.status = status

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)


class HealthInfo(object):

    """FOSSology server health info.

    Represents the health endpoint of FOSSology API.

    :param status: the overall status of the API service (OK, WARN, ERROR)
    :param scheduler: the status of the FOSSology scheduler
    :param db: the status of the FOSSology DB
    :type status: string
    :type scheduler: Status object
    :type db: Status object
    :type kwargs: key word argument
    """

    def __init__(self, status, scheduler, db, **kwargs):
        self.status = status
        self.scheduler = Status.from_json(scheduler)
        self.db = Status.from_json(db)

    def __str__(self):
        return f"FOSSology server status is: {self.status} (Scheduler: {self.scheduler.status} - DB: {self.db.status})"

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)


class SearchResult(object):

    """Search result.

    Represents a search response from FOSSology API.

    :param upload: upload where the searched file has been found
    :param uploadTreeId: id of the upload tree
    :param filename: filename of the tree item
    :param kwargs: handle any other job information provided by the fossology instance
    :type upload: Upload
    :type uploadTreeId: int
    :type filename: string
    :type kwargs: key word argument
    """

    def __init__(self, upload, uploadTreeId, filename, **kwargs):
        self.upload = Upload.from_json(upload)
        self.uploadTreeId = uploadTreeId
        self.filename = filename
        self.additional_info = kwargs

    def __str__(self):
        return f"File found in upload {self.upload.uploadname} ({self.uploadTreeId}): {self.filename}"

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)


class GetClearingHistory(object):

    """Clearing history.

    Represents the clearing history of a specified item.

    :param date: date of the clearing history
    :param username: username of the user who created the decision
    :param scope: scope of the clearing
    :param type: type of the clearing
    :param addedLicenses: list of license shortnames added to the decision
    :param removedLicenses: list of license shortnames removed to the decision
    :param kwargs: handle any other job information provided by the fossology instance
    :type date: string
    :type username: string
    :type scope: str
    :type type: str
    :type addedLicenses: List[str]
    :type removedLicenses: List[str]
    :type kwargs: key word argument
    """

    def __init__(
        self,
        date: str,
        username: str,
        scope: str,
        type: str,
        addedLicenses: Iterable[str],
        removedLicenses: Iterable[str],
        **kwargs,
    ):
        self.date = date
        self.username = username
        self.scope = ClearingScope(scope)
        self.type = ClearingType(type)
        self.addedLicenses = addedLicenses
        self.removedLicenses = removedLicenses
        self.additional_info = kwargs

    def __str__(self):
        return f"{self.username} changed clearing history at {self.date} in {self.scope} (type: {self.type})"

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)


class GetBulkHistory(object):

    """Bulk history.

    Represents the bulk history of a specified item.

    :param bulkId: the bulk id
    :param clearingEventId: the event id associated with the bulk
    :param text: scan reference text
    :param matched: whether matched or not
    :param tried: whether tried or not
    :param addedLicenses: list of license shortnames added to the scan
    :param removedLicenses: list of license shortnames removed to the scan
    :param kwargs: handle any other job information provided by the fossology instance
    :type bulkId: int
    :type clearingEventId: int
    :type text: str
    :type matched: bool
    :type tried: bool
    :type addedLicenses: List[str]
    :type removedLicenses: List[str]
    :type kwargs: key word argument
    """

    def __init__(
        self,
        bulkId: int,
        clearingEventId: int,
        text: str,
        matched: bool,
        tried: bool,
        addedLicenses: Iterable[str],
        removedLicenses: Iterable[str],
        **kwargs,
    ):
        self.bulkId = bulkId
        self.clearingEventId = clearingEventId
        self.text = text
        self.matched = matched
        self.tried = tried
        self.addedLicenses = addedLicenses
        self.removedLicenses = removedLicenses
        self.additional_info = kwargs

    def __str__(self):
        return f"Bulk Id {self.bulkId} associated with {self.clearingEventId} | Search for {self.text}"

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)


class GetPrevNextItem(object):
    """PrevNext item for the clearing history.

    Represents the prev-next item list for the clearing history.

    :param prevItemId: id of the previous item
    :param nextItemId: id of the next item
    :param kwargs: handle any other job information provided by the fossology instance
    :type prevItemId: int
    :type nextItemId: int
    :type kwargs: key word argument
    """

    def __init__(
        self,
        prevItemId: int,
        nextItemId: int,
        **kwargs,
    ):
        self.prevItemId = prevItemId
        self.nextItemId = nextItemId
        self.additional_info = kwargs

    def __str__(self):
        return f"Prev: {self.prevItemId} | Next: {self.nextItemId}"

    @classmethod
    def from_json(cls, json_dict):
        return cls(**json_dict)



# mypy: disable-error-code="attr-defined"
# Copyright 2019 Siemens AG
# SPDX-License-Identifier: MIT
import json
import logging

from fossology.enums import CopyrightStatus, PrevNextSelection
from fossology.exceptions import FossologyApiError
from fossology.obj import (
    FileInfo,
    GetBulkHistory,
    GetClearingHistory,
    GetPrevNextItem,
    Upload,
)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class Items:
    """Class dedicated to all "uploads...items" related endpoints"""

    def item_info(
        self,
        upload: Upload,
        item_id: int,
    ) -> FileInfo:
        """Get the info for a specific upload item

        API Endpoint: GET /uploads/{id}/item/{itemId}/info

        :param upload: the upload to get items from
        :param item_id: the id of the item
        :type upload: Upload
        :type item_id: int,
        :return: the file info for the specified item
        :rtype: FileInfo
        :raises FossologyApiError: if the REST call failed
        """
        response = self.session.get(
            f"{self.api}/uploads/{upload.id}/item/{item_id}/info"
        )

        if response.status_code == 200:
            return FileInfo.from_json(response.json())

        elif response.status_code == 404:
            description = f"Upload {upload.id} or item {item_id} not found"
            raise FossologyApiError(description, response)
        else:
            description = f"API error while getting info for item {item_id} from upload {upload.uploadname}"
            raise FossologyApiError(description, response)

    def item_copyrights(
        self,
        upload: Upload,
        item_id: int,
        status: CopyrightStatus,
    ) -> int:
        """Get the total copyrights of the mentioned upload tree ID

        API Endpoint: GET /uploads/{id}/item/{itemId}/totalcopyrights

        :param upload: the upload to get items from
        :param item_id: the id of the item
        :param status: the status of the copyrights
        :type upload: Upload
        :type item_id: int,
        :return: the total number of copyrights for the uploadtree item
        :rtype: int
        :raises FossologyApiError: if the REST call failed
        """
        response = self.session.get(
            f"{self.api}/uploads/{upload.id}/item/{item_id}/totalcopyrights?status={status.value}"
        )

        if response.status_code == 200:
            return response.json()["total_copyrights"]

        elif response.status_code == 404:
            description = f"Upload {upload.id} or item {item_id} not found"
            raise FossologyApiError(description, response)
        else:
            description = f"API error while getting total copyrights for item {item_id} from upload {upload.uploadname}."
            raise FossologyApiError(description, response)

    def get_clearing_history(
        self,
        upload: Upload,
        item_id: int,
    ) -> list[GetClearingHistory]:
        """Get the clearing history for a specific upload item

        API Endpoint: GET /uploads/{id}/item/{itemId}/clearing-history

        :param upload: the upload to get items from
        :param item_id: the id of the item with clearing decision
        :type upload: Upload
        :type item_id: int,
        :return: the clearing history for the specified item
        :rtype: List[GetClearingHistory]
        :raises FossologyApiError: if the REST call failed
        :raises AuthorizationError: if the REST call is not authorized
        """
        response = self.session.get(
            f"{self.api}/uploads/{upload.id}/item/{item_id}/clearing-history"
        )

        if response.status_code == 200:
            clearing_history = []
            for action in response.json():
                clearing_history.append(GetClearingHistory.from_json(action))
            return clearing_history

        elif response.status_code == 404:
            description = f"Upload {upload.id} or item {item_id} not found"
            raise FossologyApiError(description, response)
        else:
            description = f"API error while getting clearing history for item {item_id} from upload {upload.uploadname}."
            raise FossologyApiError(description, response)

    def get_prev_next(
        self, upload: Upload, item_id: int, selection: PrevNextSelection | None = None
    ) -> GetPrevNextItem:
        """Get the index of the previous and the next time for an upload

        API Endpoint: GET /uploads/{id}/item/{itemId}/prev-next

        :param upload: the upload to get items from
        :param item_id: the id of the item with clearing decision
        :param selection: tell Fossology server how to select prev-next item
        :type upload: Upload
        :type item_id: int
        :type selection: str
        :return: list of items for the clearing history
        :rtype: List[GetPrevNextItem]
        :raises FossologyApiError: if the REST call failed
        :raises AuthorizationError: if the REST call is not authorized
        """
        params = {}
        if selection:
            params["selection"] = selection

        response = self.session.get(
            f"{self.api}/uploads/{upload.id}/item/{item_id}/prev-next", params=params
        )

        if response.status_code == 200:
            return GetPrevNextItem.from_json(response.json())

        elif response.status_code == 404:
            description = f"Upload {upload.id} or item {item_id} not found"
            raise FossologyApiError(description, response)
        else:
            description = f"API error while getting prev-next items for {item_id} from upload {upload.uploadname}."
            raise FossologyApiError(description, response)

    def get_bulk_history(
        self,
        upload: Upload,
        item_id: int,
    ) -> list[GetBulkHistory]:
        """Get the bulk history for a specific upload item

        API Endpoint: GET /uploads/{id}/item/{itemId}/bulk-history

        :param upload: the upload to get items from
        :param item_id: the id of the item with clearing decision
        :type upload: Upload
        :type item_id: int
        :return: list of data from the bulk history
        :rtype: List[GetBulkHistory]
        :raises FossologyApiError: if the REST call failed
        :raises AuthorizationError: if the REST call is not authorized
        """
        response = self.session.get(
            f"{self.api}/uploads/{upload.id}/item/{item_id}/bulk-history"
        )

        if response.status_code == 200:
            bulk_history = []
            for item in response.json():
                bulk_history.append(GetBulkHistory.from_json(item))
            return bulk_history

        elif response.status_code == 404:
            description = f"Upload {upload.id} or item {item_id} not found"
            raise FossologyApiError(description, response)
        else:
            description = f"API error while getting bulk history for {item_id} from upload {upload.uploadname}."
            raise FossologyApiError(description, response)

    def schedule_bulk_scan(
        self,
        upload: Upload,
        item_id: int,
        spec: dict,
    ):
        """Schedule a bulk scan for a specific upload item

        API Endpoint: POST /uploads/{id}/item/{itemId}/bulk-scan

        Bulk scan specifications `spec` are added to the request body,
        following options are available:

        >>> bulk_scan_spec = {
        ...     "bulkActions": [
        ...         {
        ...             "licenseShortName": 'MIT',
        ...             "licenseText": 'License text',
        ...             "acknowledgement": 'Acknowledgment text',
        ...             "comment": 'Comment text',
        ...             "licenseAction": 'ADD', # or 'REMOVE'
        ...         }
        ...     ],
        ...     "refText": 'Reference Text',
        ...     "bulkScope": 'folder', # or upload
        ...     "forceDecision": 'false',
        ...     "ignoreIrre": 'false',
        ...     "delimiters": 'DEFAULT',
        ...     "scanOnlyFindings": 'true',
        ... }

        :param upload: the upload for the bulk scan
        :param item_id: the id of the item for the bulk scan
        :param spec: bulk scan specification
        :type upload: Upload
        :type item_id: int
        :raises FossologyApiError: if the REST call failed
        :raises AuthorizationError: if the REST call is not authorized
        """
        headers = {"Content-Type": "application/json"}
        response = self.session.post(
            f"{self.api}/uploads/{upload.id}/item/{item_id}/bulk-scan",
            headers=headers,
            data=json.dumps(spec),
        )
        if response.status_code == 201:
            logger.info(
                f"Bulk scan scheduled for upload {upload.uploadname}, item {item_id}"
            )
        elif response.status_code == 400:
            description = (
                f"Bad bulk scan request for upload {upload.id}, item {item_id}"
            )
            raise FossologyApiError(description, response)
        elif response.status_code == 404:
            description = f"Upload {upload.id} or item {item_id} not found"
            raise FossologyApiError(description, response)
        else:
            description = f"API error while scheduling bulk scan for item {item_id} from upload {upload.uploadname}."
            raise FossologyApiError(description, response)



# mypy: disable-error-code="attr-defined"
# Copyright 2019 Siemens AG
# SPDX-License-Identifier: MIT

import logging

from fossology.enums import MemberPerm
from fossology.exceptions import FossologyApiError
from fossology.obj import Group, UserGroupMember

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class Groups:
    """Class dedicated to all "groups" related endpoints"""

    def list_groups(self, deletable: bool = False) -> list[Group]:
        """Get the list of groups (accessible groups for user, all groups for admin)

        If parameter deletable is True, the method will return only deletable groups.

        API Endpoint: GET /groups

        :param deletable: wether to limit the scope only to deletable groups
        :type deletable: bool (default: False)
        :return: a list of groups
        :rtype: list()
        :raises FossologyApiError: if the REST call failed
        """
        endpoint = f"{self.api}/groups"
        endpoint += "/deletable" if deletable else ""
        response = self.session.get(endpoint)
        if response.status_code == 200:
            groups_list = []
            response_list = response.json()
            for group in response_list:
                single_group = Group.from_json(group)
                groups_list.append(single_group)
            return groups_list
        else:
            description = f"Unable to get a list of {'deletable ' if deletable else ''}groups for {self.user.name}"
            raise FossologyApiError(description, response)

    def list_group_members(self, group_id: int) -> list[UserGroupMember]:
        """Get the list of members for a given group (accessible groups for user, all groups for admin)

        API Endpoint: GET /groups/{id}/members

        :return: a list of members
        :rtype: list()
        :raises FossologyApiError: if the REST call failed
        """
        response = self.session.get(f"{self.api}/groups/{group_id}/members")
        if response.status_code == 200:
            members_list = []
            response_list = response.json()
            for member in response_list:
                single_member = UserGroupMember.from_json(member)
                members_list.append(single_member)
            return members_list
        else:
            description = f"Unable to get a list of members for group {group_id}"
            raise FossologyApiError(description, response)

    def create_group(self, name: str):
        """Create a group

        API Endpoint: POST /groups

        :param name: the name of the group
        :type name: str
        :raises FossologyApiError: if the REST call failed
        """
        headers = {"name": f"{name}"}
        response = self.session.post(f"{self.api}/groups", headers=headers)

        if response.status_code == 200:
            logger.info(f"Group '{name}' has been added")
        else:
            description = f"Group {name} already exists, failed to create group or no group name provided"
            raise FossologyApiError(description, response)

    def delete_group(self, group_id: int):
        """Create a group

        API Endpoint: DELETE /groups/{group_id}

        :param group_id: the id of the group
        :type group_id: int
        :raises FossologyApiError: if the REST call failed
        """
        response = self.session.delete(f"{self.api}/groups/{group_id}")

        if response.status_code == 202:
            logger.info(f"Group {group_id} will be deleted")
        else:
            description = f"Group {group_id} could not be deleted"
            raise FossologyApiError(description, response)

    def add_group_member(
        self, group_id: int, user_id: int, perm: MemberPerm = MemberPerm.USER
    ):
        """Add a user to a group

        API Endpoint: POST /groups/{group_id}/user/{user_id}

        :param group_id: the id of the group
        :param user_id: the id of the user
        :param perm: the permission level for the user
        :type group_id: int
        :type user_id: int
        :type perm: MemberPerm
        :raises FossologyApiError: if the REST call failed
        """
        data = dict()
        data["perm"] = perm.value
        response = self.session.post(
            f"{self.api}/groups/{group_id}/user/{user_id}", json=data
        )
        if response.status_code == 200:
            logger.info(f"User {user_id} has been added to group {group_id}.")
        elif response.status_code == 400:
            logger.info(f"User {user_id} is already a member of group {group_id}.")
        else:
            description = (
                f"An error occurred while adding user {user_id} to group {group_id}"
            )
            raise FossologyApiError(description, response)

    def delete_group_member(self, group_id: int, user_id: int):
        """Delete a user from a group

        API Endpoint: DELETE /groups/{group_id}/user/{user_id}

        :param group_id: the id of the group
        :param user_id: the id of the user
        :type group_id: int
        :type user_id: int
        :raises FossologyApiError: if the REST call failed
        """
        response = self.session.delete(f"{self.api}/groups/{group_id}/user/{user_id}")
        if response.status_code == 200:
            logger.info(f"User {user_id} will be removed from group {group_id}.")
        elif response.status_code == 400:
            description = f"Validation error while removing member {user_id} from group {group_id}."
            raise FossologyApiError(description, response)
        elif response.status_code == 404:
            description = f"Member {user_id} or group {group_id} not found."
            raise FossologyApiError(description, response)
        else:
            description = (
                f"An error occurred while deleting user {user_id} from group {group_id}"
            )
            raise FossologyApiError(description, response)
